# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Exhaustive value search (enumerating by weight of expression)."""

import collections
import keyword
import re
import sys
import timeit
import tokenize
from typing import Any, Dict, List, NamedTuple, Optional, Set, Text, Tuple, Union

from absl import logging
import numpy as np
import six
import tensorflow as tf
from tf_coder import tf_functions
from tf_coder.benchmarks import benchmark as benchmark_module
from tf_coder.datasets import collect_tensor_data
from tf_coder.models import tensor_features_model
from tf_coder.natural_language import description_handler as description_handler_module
from tf_coder.value_search import all_operations
from tf_coder.value_search import filtered_values_cache
from tf_coder.value_search import operation_base
from tf_coder.value_search import operation_filtering
from tf_coder.value_search import operation_statistics
from tf_coder.value_search import value as value_module
from tf_coder.value_search import value_search_settings as settings_module


ValuesByWeight = operation_base.ValuesByWeightDict
DescriptionHandler = description_handler_module.DescriptionHandler

Solution = NamedTuple('Solution', [
    ('value', value_module.Value),
    ('expression', Text),
    ('weight', int),
    ('time', float),
])

ValueSearchResults = NamedTuple('ValueSearchResults', [
    ('solutions', List[Solution]),
    ('total_time', float),
    ('value_set', Set[value_module.Value]),
    ('values_by_weight', ValuesByWeight),
    ('benchmark', benchmark_module.Benchmark),
    ('settings', settings_module.Settings),
    ('statistics', Optional[operation_statistics.OperationStatistics]),
])


def _suppress_warnings() -> None:
  """Suppress TensorFlow and Numpy warnings."""
  # TensorFlow will produce tons of error logging because we often apply
  # TensorFlow operations with bad arguments. Suppressing logging noticeably
  # improves performance.
  logging.set_verbosity(logging.ERROR)

  # Numpy sometimes produces warnings for overflow, etc., which can be
  # distracting.
  np.seterr(all='ignore')


def _user_inputs(inputs: Union[Dict[Text, Any], List[Any]]) -> List[Any]:
  """Takes the inputs dict or list and extracts the input tensors."""
  if isinstance(inputs, list):
    return inputs
  elif isinstance(inputs, dict):
    return list(inputs.values())
  elif isinstance(inputs, tuple):
    return list(inputs)
  else:
    raise ValueError('inputs must be a list or dict, but is {}'.format(
        type(inputs)))


def _contains_sparse(benchmark: benchmark_module.Benchmark) -> bool:
  """Returns whether the benchmark involves SparseTensors."""
  # TODO(kshi): These heuristics are okay, but we should let the user choose if
  # they want to.
  for example in benchmark.examples:
    if (isinstance(example.output, tf.SparseTensor) or
        any(isinstance(input_object, tf.SparseTensor)
            for input_object in _user_inputs(example.inputs))):
      return True
  return 'sparse' in benchmark.description.lower()


def _add_value_by_weight(values_by_weight: ValuesByWeight,
                         value: value_module.Value,
                         weight: int) -> None:
  """Adds a value of a given weight to values_by_weight."""
  if weight < len(values_by_weight):
    values_by_weight[weight][value] = value


def _constant_exists(constant: Any, constants_so_far: Set[Any]) -> bool:
  """Checks whether a constant exists already."""
  # We can't use the `in` keyword because `True in [1, 2, 3]` evaluates to True!
  # (`True == 1` evaluates to True.)
  return any(constant == existing and type(constant) is type(existing)
             for existing in constants_so_far)


def _is_valid_name(name: Text) -> bool:
  """Returns whether name is an acceptable Python identifier."""
  # Behavior is slightly different between Python versions, e.g., `await` is a
  # keyword only in PY3, and `print` is keyword only in PY2.
  if name in ['tf', 'np'] or keyword.iskeyword(name):
    return False
  if six.PY3:
    return name.isidentifier()
  else:
    return (bool(re.match(tokenize.Name + '$', name)) and
            name not in ['True', 'False', 'None'])


def _input_names_to_objects(
    inputs_collection: Union[List[Any], Dict[Text, Any]]) -> Dict[Text, Any]:
  """Returns a mapping from input names to objects, also validating names."""
  if isinstance(inputs_collection, (list, tuple)):
    input_names_to_objects = collections.OrderedDict(
        ('in' + str(i + 1), input_object)
        for i, input_object in enumerate(inputs_collection))
  elif isinstance(inputs_collection, dict):
    for name in inputs_collection:
      if not isinstance(name, six.string_types):
        raise ValueError('The input name {!r} must be a string.'.format(name))
      if not _is_valid_name(name):
        raise ValueError('The input name {!r} is not a valid Python identifier.'
                         .format(name))
    input_names_to_objects = inputs_collection
  else:
    raise ValueError('The collection of inputs has the wrong format. It can be '
                     'a list of input objects, or a dict mapping string names '
                     'to input objects.')
  return input_names_to_objects


def _add_constants_and_inputs_and_print(
    values_by_weight: ValuesByWeight,
    benchmark: benchmark_module.Benchmark,
    output_value: value_module.OutputValue,
    constant_operation: operation_base.Operation,
    settings: settings_module.Settings) -> None:
  """Adds constant/input Values to values_by_weight, and prints to stdout."""
  # Conceptually this is a set, but it's actually a list so that constants can
  # be printed in the same order they are chosen by the heuristics. The reduced
  # efficiency of membership-checking is not a big deal because we have few
  # constants.
  constants_so_far = set()
  constants_to_print = []

  # User-provided constants.
  for c in benchmark.constants:
    if not _constant_exists(c, constants_so_far):
      constant_value = value_module.ConstantValue(c)
      weight = tf_functions.PROVIDED_CONSTANT_WEIGHT
      _add_value_by_weight(values_by_weight, constant_value, weight)
      constants_so_far.add(c)
      constants_to_print.append(c)

  # Add inputs, while computing some info for extra constants later.
  max_input_tensor_rank = 0
  dimension_lengths = set()
  input_names_to_objects = _input_names_to_objects(benchmark.examples[0].inputs)
  for name, input_object in input_names_to_objects.items():
    input_value = value_module.InputValue(input_object, name)
    if input_value.is_tensor:
      max_input_tensor_rank = max(max_input_tensor_rank, len(input_value.shape))
      dimension_lengths.update(input_value.shape)
    if input_value.is_primitive and constant_operation is not None:
      scalar_tensor_value = constant_operation.apply([input_value], settings)
      _add_value_by_weight(values_by_weight, scalar_tensor_value,
                           tf_functions.PRIMITIVE_INPUT_AS_TENSOR_WEIGHT)

    _add_value_by_weight(values_by_weight, input_value,
                         tf_functions.INPUT_VARIABLE_WEIGHT)
    if input_value.is_primitive:
      constants_so_far.add(input_value.value)

    print("Input '{}':\n{!s}\n".format(name, input_value.value))

  if output_value.shape is not None:
    dimension_lengths.update(output_value.shape)

  # Always include these as constants.
  common_constants = [0, 1, -1, True, False]
  # Also include 2, 3, ..., max_example_input_tensor_rank - 1 when applicable.
  axis_constants = list(range(2, max_input_tensor_rank))
  # Also include dimension lengths of input and output tensors.
  shape_constants = sorted(dimension_lengths)

  constant_weight_pairs = (
      [(c, tf_functions.COMMON_CONSTANT_WEIGHT) for c in common_constants] +
      [(c, tf_functions.AXIS_CONSTANT_WEIGHT) for c in axis_constants] +
      [(c, tf_functions.SHAPE_CONSTANT_WEIGHT) for c in shape_constants])

  for constant, weight in constant_weight_pairs:
    if not _constant_exists(constant, constants_so_far):
      constant_value = value_module.ConstantValue(constant)
      _add_value_by_weight(values_by_weight, constant_value, weight)
      constants_so_far.add(constant)
      constants_to_print.append(constant)

  # DTypes for casting.
  for dtype, weight in tf_functions.CONSTANT_DTYPES_AND_WEIGHTS.items():
    dtype_value = value_module.ConstantValue(dtype)
    _add_value_by_weight(values_by_weight, dtype_value, weight)

  if output_value.shape:
    # Add the output shape as a constant.
    shape_tuple = tuple(output_value.shape)
    shape_tuple_value = value_module.ConstantValue(shape_tuple)
    weight = tf_functions.OUTPUT_SHAPE_TUPLE_WEIGHT
    _add_value_by_weight(values_by_weight, shape_tuple_value, weight)
    # Don't add shape_tuple to constants_to_print, because printing it out could
    # be confusing to users.

  # Only for experiments in the PLDI paper.
  if settings.paper_experiments.uniform_weights:
    # Count the number of values.
    num_values = sum(len(values_with_weight)
                     for values_with_weight in values_by_weight)
    # Take all values and put them in the collection for weight 1.
    for weight in range(2, len(values_by_weight)):
      for heavy_value in values_by_weight[weight]:
        values_by_weight[1][heavy_value] = heavy_value
      values_by_weight[weight].clear()
    # Make sure we did it right.
    for weight, values_with_weight in enumerate(values_by_weight):
      assert len(values_with_weight) == (num_values if weight == 1 else 0)

  print('Output:\n{!s}\n'.format(output_value.value))
  print('Constants: {!r}\n'.format(constants_to_print))
  if benchmark.description:
    print('Description: {}\n'.format(benchmark.description))
  print('Searching...\n')
  sys.stdout.flush()  # Flush so the inputs/output appear in Colab immediately.


def _check_solution(expression: Text,
                    used_input_names: Set[Text],
                    benchmark: benchmark_module.Benchmark,
                    settings: settings_module.Settings) -> bool:
  """Checks that the solution is good."""
  del expression  # Unused for now.

  if settings.require_all_inputs_used:
    if len(used_input_names) < len(benchmark.examples[0].inputs):
      return False
  elif settings.require_one_input_used:
    if not used_input_names:
      return False

  # TODO(kshi): Check that the solution works (floating-point errors may
  # accumulate beyond an acceptable threshold).
  return True


def _record_solutions(value: value_module.Value,
                      weight: int,
                      start_time: float,
                      solutions: List[Solution],
                      solution_expression_set: Set[Text],
                      benchmark: benchmark_module.Benchmark,
                      settings: settings_module.Settings) -> None:
  """Records new solutions in the `solutions` list."""
  reconstructions = value.reconstruct_all_expressions_with_input_names()
  this_solution_time = timeit.default_timer() - start_time
  for expression, used_input_names in reconstructions:
    if expression in solution_expression_set:
      continue
    if not _check_solution(expression, used_input_names, benchmark, settings):
      if settings.printing.bad_solutions:
        print('Bad solution: {}'.format(expression))
      continue
    solution_expression_set.add(expression)
    solutions.append(Solution(value=value, expression=expression,
                              weight=weight, time=this_solution_time))
    print('Found solution: {}'.format(expression))
    # Flush so the solutions appear in Colab immediately.
    sys.stdout.flush()
    if len(solutions) >= settings.max_solutions:
      break


def _find_solutions(
    benchmark: benchmark_module.Benchmark,
    operations: List[operation_base.Operation],
    start_time: float,
    settings: settings_module.Settings
) -> Tuple[List[Solution], Set[value_module.Value], ValuesByWeight,
           Optional[operation_statistics.OperationStatistics]]:
  """Helper, returning (solutions, value_set, values_by_weight, statistics)."""
  timeout_reached = False
  end_time = start_time + settings.timeout

  only_minimal_solutions = settings.only_minimal_solutions
  if settings.max_solutions == 1:
    # If we only want one solution, it will be minimal.
    only_minimal_solutions = True

  # An object to track statistics, if requested.
  statistics = (operation_statistics.OperationStatistics()
                if settings.printing.statistics
                else None)

  # A list of Solution namedtuples.
  solutions = []

  # A set of string solution expressions (don't return duplicate solutions).
  solution_expression_set = set()

  # The output value to search for.
  output_value = value_module.OutputValue(benchmark.examples[0].output)

  # A list of OrderedDicts mapping Value objects to themselves. The i-th
  # OrderedDict contains all Value objects of weight i.
  values_by_weight = [collections.OrderedDict()
                      for _ in range(settings.max_weight + 1)]

  # Find and cache the cast and constant operations for use later.
  cast_operation = None
  constant_operation = None
  for operation in operations:
    if operation.name == tf_functions.CAST_OPERATION_NAME:
      cast_operation = operation
    elif operation.name == tf_functions.CONSTANT_OPERATION_NAME:
      constant_operation = operation
  # Create the output dtype value for use later.
  dtype_value = value_module.ConstantValue(output_value.dtype)

  # Populate values_by_weight with inputs and constants. This also prints
  # inputs/output/constants to stdout.
  _add_constants_and_inputs_and_print(
      values_by_weight, benchmark, output_value, constant_operation, settings)

  # A set storing all values found so far.
  value_set = set().union(*values_by_weight)

  # Check if the output equals an input or constant.
  for weight, values in enumerate(values_by_weight):
    if output_value in values:
      matching_value = values[output_value]
      print('WARNING: The output is the same as the input or constant `{}`. '
            'Check that your input-output example is correct.\n'.format(
                matching_value.reconstruct_expression()))
      possible_first_solution = not solutions
      _record_solutions(matching_value, weight, start_time, solutions,
                        solution_expression_set, benchmark, settings)
      if possible_first_solution and solutions:
        end_time = min(
            end_time,
            timeit.default_timer() + settings.max_extra_solutions_time)
      if len(solutions) >= settings.max_solutions:
        return solutions, value_set, values_by_weight, statistics

  filter_cache = filtered_values_cache.FilteredValuesCache()

  # Value search by weight.
  for weight in range(1, settings.max_weight + 1):
    if settings.printing.progress:
      print('Searching weight {}...'.format(weight))

    # Values with the current weight. This might already include leaf values.
    new_values = values_by_weight[weight]

    for operation in operations:
      for value in operation.enumerate_values_with_weight(
          target_weight=weight,
          values_by_weight=values_by_weight,
          filter_cache=filter_cache,
          end_time=end_time,
          settings=settings,
          statistics=statistics):

        if value not in value_set:
          # This value has never been seen before, or it's the desired output.
          if settings.printing.verbose:
            expression = value.reconstruct_expression()
            print('{} produces:\n{}'.format(expression, value))

          if value == output_value:
            possible_first_solution = not solutions
            # Found solution(s), but some may be bad.
            _record_solutions(value, weight, start_time, solutions,
                              solution_expression_set, benchmark, settings)
            if possible_first_solution and solutions:
              end_time = min(
                  end_time,
                  timeit.default_timer() + settings.max_extra_solutions_time)
            if len(solutions) >= settings.max_solutions:
              return solutions, value_set, values_by_weight, statistics
          else:
            # Only store the value if it isn't a solution. Otherwise, we'll get
            # lots of "almost duplicate" solutions, e.g., by adding 0.
            new_values[value] = value
            # We should never add output_value (or anything equal) to value_set
            # so that we can continue finding other solutions.
            value_set.add(value)
        else:  # This value has been seen before.
          if value in new_values:
            # The value was already computed differently with this weight.
            original_value = new_values[value]
            if isinstance(original_value, value_module.OperationValue):
              # Only merge reconstructions if this was originally an
              # OperationValue. (It could be a ConstantValue instead.)
              operation_value = original_value   # type: value_module.OperationValue
              operation_value.merge_reconstructions(value)
          elif not only_minimal_solutions:
            # If we want non-minimal solutions, we need to store the value even
            # if we have already seen that value with a smaller weight.
            new_values[value] = value

      if timeit.default_timer() > end_time:
        timeout_reached = True
        # Don't return immediately; still try to cast new values because this is
        # relatively quick.
        break

    # Try casting new values to the output dtype if this has a chance of being
    # a correct solution.
    for new_value in new_values:
      if (cast_operation is not None and
          new_value.shape == output_value.shape and
          new_value.dtype != output_value.dtype and
          operation_filtering.is_castable(new_value, dtype_value)):
        casted_value = cast_operation.apply([new_value, dtype_value], settings)
        if casted_value == output_value:
          possible_first_solution = not solutions
          # Found solution(s), but some may be bad.
          _record_solutions(casted_value, weight, start_time, solutions,
                            solution_expression_set, benchmark, settings)
          if possible_first_solution and solutions:
            end_time = min(
                end_time,
                timeit.default_timer() + settings.max_extra_solutions_time)
          if len(solutions) >= settings.max_solutions:
            return solutions, value_set, values_by_weight, statistics

    if settings.printing.progress:
      print('Found {} distinct values of weight {}, or {} total.'.format(
          len(new_values), weight, len(value_set)))
    if only_minimal_solutions and solutions:
      return solutions, value_set, values_by_weight, statistics
    if timeout_reached:
      break

  return solutions, value_set, values_by_weight, statistics


def operation_multipliers_from_tensor_model(
    benchmark: benchmark_module.Benchmark,
    tensor_model: tensor_features_model.Model,
    tensor_config: Dict[Text, Any],
    settings: settings_module.Settings) -> Dict[Text, float]:
  """Runs the tensor features model to get operation weight multipliers."""
  # TODO(kshi): Duplicate creation of InputValue and OutputValue objects.
  inputs = [value_module.InputValue(user_input, '')
            for user_input in _user_inputs(benchmark.examples[0].inputs)]
  output = value_module.OutputValue(benchmark.examples[0].output)
  example_protos = collect_tensor_data.create_tf_examples(
      io_example=collect_tensor_data.IOExample(
          expression='',
          input_values=inputs,
          output_value=output,
          num_inputs=benchmark.num_inputs,
          operations=[]),
      max_num_inputs=tensor_config['max_num_inputs'],
      permute_inputs=False)
  if example_protos:
    example_proto = example_protos[0]
    result = tensor_features_model.eval_single_example(
        tensor_model, example_proto.SerializeToString())
    operation_ranks = tf.argsort(tf.argsort(tf.squeeze(result.operation_logits),
                                            stable=True))
    operation_probabilities = tf.squeeze(tf.sigmoid(result.operation_logits))
    operation_names = tensor_config['operation_names']
    assert len(operation_names) == len(operation_probabilities)

    multipliers = {}
    for index in tf.where(operation_probabilities
                          >= settings.tensor_model.prioritize_threshold):
      chosen_op_name = operation_names[int(index)]
      if settings.printing.prioritized_operations:
        print('Tensor features model prioritized {}, p={}'.format(
            chosen_op_name, operation_probabilities[int(index)]))
      multipliers[chosen_op_name] = settings.tensor_model.prioritize_multiplier
    for index in tf.where(operation_probabilities
                          <= settings.tensor_model.deprioritize_threshold):
      int_index = int(index)
      chosen_op_name = operation_names[int_index]
      if (int(operation_ranks[int_index])
          >= settings.tensor_model.max_deprioritized):
        continue
      if settings.printing.deprioritized_operations:
        print('Tensor features model deprioritized {}, p={}, logit={}'.format(
            chosen_op_name, operation_probabilities[int_index],
            result.operation_logits[0][int_index]))
      multipliers[chosen_op_name] = (
          settings.tensor_model.deprioritize_multiplier)
    return multipliers
  else:
    logging.error('Failed to create example for inputs %s and output %s',
                  inputs, output)
    return {}


def _combine_multipliers(first: Dict[Text, float],
                         second: Dict[Text, float]) -> Dict[Text, float]:
  """Combines operation weight multiplier dicts. Modifies the first dict."""
  for name in second:
    first[name] = first.get(name, 1.0) * second[name]
  return first


def get_reweighted_operations(
    benchmark: benchmark_module.Benchmark,
    settings: settings_module.Settings,
    description_handler: Optional[DescriptionHandler] = None,
    tensor_model: Optional[tensor_features_model.Model] = None,
    tensor_config: Optional[Dict[Text, Any]] = None,
) -> List[operation_base.Operation]:
  """Returns a list of operations with correct weights for the problem."""
  include_sparse_operations = (
      not settings.operations.limit_sparse_operations or
      _contains_sparse(benchmark))
  operations = all_operations.get_operations(
      include_sparse_operations=include_sparse_operations)

  operation_names = [op.name for op in operations]
  if len(operation_names) != len(set(operation_names)):
    raise ValueError('Operation names were not unique.')

  if settings.paper_experiments.uniform_weights:
    # Only for experiments in the PLDI paper.
    for operation in operations:
      operation.weight = 1
    return operations

  multipliers = {}
  if description_handler and benchmark.description:
    multipliers = _combine_multipliers(
        multipliers,
        description_handler.get_operation_multipliers(benchmark, settings))
  if tensor_model is not None and tensor_config is not None:
    multipliers = _combine_multipliers(
        multipliers,
        operation_multipliers_from_tensor_model(benchmark, tensor_model,
                                                tensor_config, settings))
  for operation in operations:
    operation.weight = max(
        1, int(round(operation.weight * multipliers.get(operation.name, 1))))

  return operations


def run_value_search(
    benchmark: benchmark_module.Benchmark,
    settings: settings_module.Settings,
    description_handler: Optional[DescriptionHandler] = None,
    tensor_model: Optional[tensor_features_model.Model] = None,
    tensor_config: Optional[Dict[Text, Any]] = None) -> ValueSearchResults:
  """Performs value search, iterating by the expression weight.

  Starts with the constants and user-provided inputs, and applies the given
  operations, for a given number of iterations. An expression's "weight" is the
  number of nodes in the expression tree.

  Args:
    benchmark: The Benchmark containing input-output examples and constants.
    settings: A Settings object containing settings for this search.
    description_handler: A DescriptionHandler that scores operations based on
      the benchmark's description.
    tensor_model: The tensor features model to use, already restored from a
      checkpoint. If None, do not run the model.
    tensor_config: The config to use with the tensor features model.

  Returns:
    A ValueSearchResults namedtuple.

  Raises:
    ValueError: If max_weight is too large to be reasonable.
  """
  _suppress_warnings()
  if len(benchmark.examples) > 1:
    print('Warning: for now, value search only uses a single example.')

  start_time = timeit.default_timer()

  operations = get_reweighted_operations(
      benchmark,
      settings,
      description_handler=description_handler,
      tensor_model=tensor_model,
      tensor_config=tensor_config)

  solutions, value_set, values_by_weight, statistics = _find_solutions(
      benchmark=benchmark,
      operations=operations,
      start_time=start_time,
      settings=settings)

  total_time = timeit.default_timer() - start_time

  if solutions:
    print()
    print('Solution was found in {:.1f} seconds:\n{}'.format(
        solutions[0].time, solutions[0].expression))
    if settings.max_solutions != 1:
      print('Found {} solution(s) in {:.1f} seconds total.'.format(
          len(solutions), total_time))
  else:
    print('Could not find solution within {} seconds.'.format(
        min(settings.timeout, total_time)))
  sys.stdout.flush()

  return ValueSearchResults(
      solutions=solutions,
      total_time=total_time,
      value_set=value_set,
      values_by_weight=values_by_weight,
      benchmark=benchmark,
      settings=settings,
      statistics=statistics)


def run_value_search_from_example(
    inputs: Union[List[Any], Dict[Text, Any]],
    output: Any,
    settings: Optional[settings_module.Settings] = None,
    **kwargs) -> ValueSearchResults:
  """Performs value search for a single user-provided input-output example.

  Args:
    inputs: A list of inputs, or a dict mapping input names to inputs.
    output: The corresponding desired output.
    settings: An optional Settings object to use, or None to use defaults.
    **kwargs: The kwarg 'constants' can be used to specify a list of constants,
      and 'description' can be used to provide a natural language description of
      the task. Other arguments are passed directly to run_value_search().

  Returns:
    A ValueSearchResults namedtuple.
  """
  if settings is None:
    settings = settings_module.default_settings()
  constants = kwargs.pop('constants', None)
  description = kwargs.pop('description', None)
  source = kwargs.pop('source', 'From user-provided example.')
  benchmark = benchmark_module.Benchmark(
      examples=[benchmark_module.Example(inputs, output)],
      constants=constants,  # Will turn into empty list if constants=None.
      description=description,  # Will turn into '' if description=None.
      source=source)

  return run_value_search(benchmark, settings, **kwargs)



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""TF-IDF for similarity between task descriptions and operation docstrings."""

import re
from typing import Dict, List, Optional, Set, Text

from sklearn.feature_extraction import text as sklearn_text
from sklearn.metrics import pairwise as sklearn_pairwise
from tf_coder.benchmarks import benchmark as benchmark_module
from tf_coder.natural_language import description_handler
from tf_coder.value_search import operation_base
from tf_coder.value_search import value_search_settings as settings_module


class TfidfDescriptionHandler(description_handler.DescriptionHandler):
  """Computes similarity between task descriptions and operation docstrings."""

  def __init__(self,
               operations: Optional[List[operation_base.Operation]] = None,
               max_num_prioritized: int = 5,
               min_tfidf_score: float = 0.15,
               multiplier: float = 0.75):
    """Creates a TfidfDescriptionHandler that can score the given operations.

    Args:
      operations: A list of operations that the scorer should handle. Exposed
        for testing.
      max_num_prioritized: The maximum number of operations to prioritize (these
        are assigned a delta of -1).
      min_tfidf_score: The minimum TF-IDF score for an operation to be
        prioritized.
      multiplier: The multiplier applied to an operation's weight if it is
        prioritized.

    Raises:
      ValueError: If there are duplicate operation names.
    """
    super(TfidfDescriptionHandler, self).__init__(operations)
    self.max_num_prioritized = max_num_prioritized
    self.min_tfidf_score = min_tfidf_score
    self.multiplier = multiplier

    docstrings = [operation.metadata.docstring.lower()
                  for operation in self.operations]
    self._vectorizer = sklearn_text.TfidfVectorizer(stop_words='english')
    self._term_document_matrix = self._vectorizer.fit_transform(docstrings)

  def score_description(self, description: Text) -> Dict[Text, float]:
    """Returns a map from operation names to their similarity to `description`.

    Similarity scores will be in the range [0, 1], with higher scores indicating
    more relevance.

    Args:
      description: An English description of a task.
    """
    description_vector = self._vectorizer.transform([description.lower()])
    scores = sklearn_pairwise.cosine_similarity(
        description_vector, self._term_document_matrix).flatten()
    return dict(zip(self.all_names, scores))

  def _get_listed_operations(self, description: Text) -> Set[Text]:
    """Returns operations listed explicitly in the description.

    This handles the following advice from Colab:
    > If you know of a TensorFlow operation (e.g., tf.reduce_max) that is
      relevant, include its name (e.g., "tf.reduce_max") anywhere in the
      description. This will lead TF-Coder to prioritize that operation.

    We can be a bit more lenient and allow "reduce_max" anywhere in the
    description (without the "tf."). This helps in situations where there are
    multiple aliases, e.g., `tf.add` and `tf.math.add`, so the user doesn't
    have to exactly match the alias we use.

    Note that having "gather_nd" in the description does NOT lead to
    priotizing `tf.gather`, because we match on word boundaries.

    Also note that this may lead to prioritizing more operations than
    `self.max_num_prioritized`.

    Args:
      description: An English description of a task.
    """
    description = description.lower()
    description_words = set(re.findall(r'\w+', description))
    listed_set = set()
    for operation_name in self.all_names:
      if operation_name.startswith('tf.') and '(' in operation_name:
        paren_index = operation_name.index('(')
        dot_index = operation_name[:paren_index].rindex('.')
        name_to_find = operation_name[dot_index + 1 : paren_index]
      else:
        name_to_find = operation_name  # For Python operations.
      if name_to_find.lower() in description_words:
        listed_set.add(operation_name)
    return listed_set

  def get_operation_multipliers(
      self, benchmark: benchmark_module.Benchmark,
      settings: settings_module.Settings) -> Dict[Text, float]:
    """See base class."""
    description = benchmark.description.lower()
    scores = self.score_description(description)
    # Sorted in order of decreasing TF-IDF score.
    sorted_names = sorted(self.all_names,
                          key=lambda name: scores[name], reverse=True)
    prioritized_set = set(name
                          for name in sorted_names[:self.max_num_prioritized]
                          if scores[name] >= self.min_tfidf_score)

    prioritized_set.update(self._get_listed_operations(description))

    if settings.printing.prioritized_operations:
      for name in prioritized_set:
        print('TF-IDF handler prioritized {}, score={:.3f}'.format(
            name, scores[name]))

    return dict((name, self.multiplier) for name in prioritized_set)

  def __repr__(self) -> Text:
    """See base class."""
    return ('{}(max_num_prioritized={}, min_tfidf_score={}, multiplier={})'
            .format(self.__class__.__name__,
                    self.max_num_prioritized,
                    self.min_tfidf_score,
                    self.multiplier))



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Defines the base Operation class for value search."""

import abc
import itertools
import sys
import timeit
import typing
from typing import Callable, Dict, List, Optional, Sequence, Set, Text, Tuple, Union

import six
from tf_coder import tf_coder_utils
from tf_coder import tf_functions
from tf_coder.value_search import filtered_values_cache
from tf_coder.value_search import operation_statistics
from tf_coder.value_search import value
from tf_coder.value_search import value_search_settings as settings_module

################################################################################
# Type aliases.

# The i-th element contains all Value objects of weight i, mapped to themselves.
ValuesByWeightDict = List[Dict[value.Value, value.Value]]

# The i-th element is an iterable of all Value objects of weight i.
ValuesByWeightIterable = List[Union[List[value.Value],
                                    Dict[value.Value, value.Value]]]

# The i-th inner list contains Value objects that are candidates for argument i.
ArgOptionsType = List[List[value.Value]]

# The i-th Value is used as the i-th argument for an Operation application.
ArgValuesType = Sequence[value.Value]

# An optional filter function that is applied to single Value objects. If None,
# it is treated as a function that always returns True (all Value objects are
# allowed).
ValueFilterType = Optional[Callable[[value.Value], bool]]

# An optional filter function that is applied to a list of argument values. If
# None, it is treated as a function that always returns True (all argument lists
# are allowed).
ApplyFilterType = Optional[Callable[[ArgValuesType], bool]]

################################################################################


OperationMetadata = typing.NamedTuple(
    'OperationMetadata', [('docstring', Text)])


@six.add_metaclass(abc.ABCMeta)
class Operation(object):
  """An operation that can be applied to a constant number of arguments.

  Arguments are always ordered, and subclasses can choose their own conventions
  for this ordering. The operation must be deterministic, must not have
  side-effects, and must not modify its arguments.

  Attributes:
    name: A unique name for this operation.
    num_args: The number of arguments required by this Operation.
    weight: The weight of this node in the AST.
    metadata: Metadata for this Operation.

    _value_filters_list: A list of lists of filter functions.

      Each inner list has length num_args and contains a filter function for
      each argument, where the i-th filter function takes a Value and returns
      whether that Value should be an option for the i-th argument. Any filter
      function can be None, which means all values should be options for that
      argument.

      The outer list can have multiple lists of filter functions, where each
      inner list describes one class of valid argument values.

      The value_filters_list attribute can also be None, in which case all
      values should be options for all arguments.

    _apply_filter: A filter function that takes a list of Value objects of
      length num_args (the arguments to a potential application of this
      Operation), and returns whether those Value objects are compatible (i.e.,
      whether the operation should be applied). If None, the operation is always
      applied.

    _name_cache: A cached copy of this Operation's name.
  """

  def __init__(self, num_args: int, weight: int,
               metadata: OperationMetadata) -> None:
    """Initializes an Operation."""
    self.num_args = num_args
    self.weight = weight
    self.metadata = metadata
    self._value_filters_list = None  # type: Optional[List[List[ValueFilterType]]]
    self._apply_filter = None  # type: ApplyFilterType
    self._name_cache = None

  @property
  def name(self) -> Text:
    """The (cached) name of the operation."""
    if self._name_cache is not None:
      return self._name_cache
    self._name_cache = self._compute_name()
    return self._name_cache

  def _compute_name(self) -> Text:
    """Computes a name for this operation."""
    return self.__class__.__name__

  def add_value_filters(self, value_filters: List[ValueFilterType]) -> None:
    """Adds the given value filters to the value_filters_list attribute.

    Args:
      value_filters: A list of filter functions, one per argument, where the
        i-th filter function takes a Value and returns whether it should be
        an option for argument i.

    Raises:
      ValueError: If the list of filter functions has the wrong length.
    """
    if len(value_filters) != self.num_args:
      raise ValueError('value_filters must contain one filter per argument.')
    if self._value_filters_list is None:
      self._value_filters_list = []
    self._value_filters_list.append(value_filters)

  def set_apply_filter(self, apply_filter: ApplyFilterType) -> None:
    """Sets the given apply_filter."""
    self._apply_filter = apply_filter

  @abc.abstractmethod
  def apply(self,
            arg_values: ArgValuesType,
            settings: settings_module.Settings) -> Optional[value.Value]:
    """Applies this Operation to a list of arguments (Value objects).

    Args:
      arg_values: A list of Value objects representing the arguments.
      settings: A Settings object storing settings for this search.

    Returns:
      A Value object representing the result if successful, or None if the
      operation raises an exception.
    """

  def _enumerate_values(
      self,
      arg_options: ArgOptionsType,
      end_time: float,
      settings: settings_module.Settings,
      statistics: Optional[operation_statistics.OperationStatistics] = None,
  ) -> List[value.Value]:
    """Enumerates values that are created from multiple choices of arguments.

    Args:
      arg_options: A list of lists of Value objects, where the i-th list
        contains the possible Value objects for the i-th argument.
      end_time: A timeit.default_timer() cutoff where this should timeout.
      settings: A Settings object storing settings for this search.
      statistics: An optional OperationStatistics object to track statistics
        during this function's execution.

    Returns:
      A list of Value objects, one for every successful application of the
      operation.
    """
    results = []  # type: List[value.Value]
    apply_count = 0
    apply_successes = 0
    start_time = timeit.default_timer()

    for i, arg_values in enumerate(itertools.product(*arg_options)):
      # Check for timeout periodically.
      if i % 1000 == 0 and timeit.default_timer() > end_time:
        break

      # Skipping filtering is only used for experiments in the PLDI paper.
      if not (settings.paper_experiments.skip_filtering and
              self.name not in tf_functions.REQUIRES_FILTERING):
        # _apply_filter is either None or callable.
        if (self._apply_filter is not None and
            not self._apply_filter(arg_values)):  # pylint: disable=not-callable
          continue

      if settings.printing.all_apply:
        print('Applying {} on arguments: {}'.format(
            self.name,
            [arg_value.reconstruct_expression() for arg_value in arg_values]))
        # Print the output immediately so it isn't swallowed by a stacktrace.
        sys.stdout.flush()
      maybe_value = self.apply(arg_values, settings)
      apply_count += 1
      if maybe_value is not None:
        yes_value = maybe_value  # type: value.Value
        apply_successes += 1
        results.append(yes_value)

    elapsed_time = timeit.default_timer() - start_time
    if statistics:
      statistics.update(operation_name=self.name,
                        count=apply_count,
                        successes=apply_successes,
                        time=elapsed_time)
    return results

  def enumerate_values_with_weight(
      self,
      target_weight: int,
      values_by_weight: ValuesByWeightDict,
      filter_cache: filtered_values_cache.FilteredValuesCache,
      end_time: float,
      settings: settings_module.Settings,
      statistics: Optional[operation_statistics.OperationStatistics] = None
  ) -> List[value.Value]:
    """Enumerates values with a given target weight.

    Args:
      target_weight: The desired weight of resulting values.
      values_by_weight: A collection of Values organized by their weight.
      filter_cache: The FilteredValuesCache object used during this search.
      end_time: A timeit.default_timer() cutoff where this should timeout.
      settings: A Settings object storing settings for this search.
      statistics: An optional OperationStatistics object to track statistics
        during this function's execution.

    Returns:
      A list of Value objects of the specified weight.
    """
    num_args = self.num_args
    if num_args == 0:
      return []  # An operation with no arguments can't have variable weight.
    if target_weight - self.weight - num_args < 0:
      return []  # Too many arguments for this weight.

    results = []  # type: List[value.Value]

    for value_filters in self._value_filters_list:
      assert len(value_filters) == num_args

      # Enumerate ways of partitioning (target_weight - self.weight) into
      # (num_args) positive pieces.
      # Equivalently, partition (target_weight - self.weight - num_args) into
      # (num_args) nonnegative pieces.
      arg_options_list = []  # type: List[ArgOptionsType]
      for partition in tf_coder_utils.generate_partitions(
          target_weight - self.weight - num_args,
          num_args):  # type: Tuple[int, ...]  # pytype: disable=annotation-type-mismatch
        if (settings.paper_experiments.skip_filtering and
            self.name not in tf_functions.REQUIRES_FILTERING):
          # Only for experiments in the PLDI paper.
          arg_options = [
              values_by_weight[weight_minus_1 + 1]
              for arg, weight_minus_1 in enumerate(partition)
          ]  # type: ArgOptionsType  # pytype: disable=annotation-type-mismatch
        else:
          arg_options = [
              filter_cache.filter_values(value_filters[arg], weight_minus_1 + 1,
                                         values_by_weight[weight_minus_1 + 1])
              for arg, weight_minus_1 in enumerate(partition)
          ]  # type: ArgOptionsType
        arg_options_list.append(arg_options)

      for arg_options in arg_options_list:
        results.extend(self._enumerate_values(arg_options, end_time, settings,
                                              statistics))
    return results

  def reconstruct_expression(self, arg_values: ArgValuesType,
                             use_cache=True) -> Text:
    """Returns an expression for this operation applied to the given arguments.

    This can be slow and should not be called in a tight loop.

    Args:
      arg_values: A list of Value objects representing the arguments' values.
      use_cache: If True, the reconstruction may be looked up from a cache. If
        False, the reconstruction will be recomputed on each call.

    Returns:
      A string representation of the code expression.
    """
    arg_strings = [arg_value.reconstruct_expression(use_cache=use_cache)
                   for arg_value in arg_values]
    return self.reconstruct_expression_from_strings(arg_strings)

  def reconstruct_expression_with_input_names(
      self, arg_values: ArgValuesType) -> Tuple[Text, Set[Text]]:
    """Returns an expression for this operation and the used input names."""
    arg_strings_list, input_names_list = zip(
        *[arg_value.reconstruct_expression_with_input_names()
          for arg_value in arg_values])
    return (self.reconstruct_expression_from_strings(arg_strings_list),
            set.union(*input_names_list))

  @abc.abstractmethod
  def reconstruct_expression_from_strings(self,
                                          arg_strings: List[Text]) -> Text:
    """Returns an expression for this operation applied to the given arguments.

    This can be slow and should not be called in a tight loop.

    Args:
      arg_strings: A list of strings representing the arguments'
        reconstructions.

    Returns:
      A string representation of the code expression.
    """



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Utilities for operation filtering."""

import functools
import math
import operator
from typing import Any, Tuple, Type

import tensorflow as tf
from tf_coder import filter_group
from tf_coder import tensor_limits as limits
from tf_coder import tf_coder_utils
from tf_coder.value_search import value as value_module


@functools.lru_cache(maxsize=None)
def get_type_filter(desired_type):
  """Returns a value filter that only keeps values of the given type."""
  return lambda arg_value: arg_value.type is desired_type


@functools.lru_cache(maxsize=None)
def get_types_filter(desired_types: Tuple[Type[Any], ...]):
  """Returns a value filter that only keeps values with the given types."""
  return lambda arg_value: arg_value.type in desired_types


@functools.lru_cache(maxsize=None)
def get_dtype_filter(dtype):
  """Returns a value filter that only keeps tensor values of the given dtype."""
  if not isinstance(dtype, tf.DType):
    raise TypeError('dtype must be a tf.DType.')
  return lambda arg_value: arg_value.dtype is dtype


@functools.lru_cache(maxsize=None)
def get_tensor_min_rank_filter(rank):
  """Returns a value filter that only keeps tensors of high enough rank."""
  return lambda arg_value: arg_value.is_tensor and len(arg_value.shape) >= rank


def _check_tensor_finite(tensor):
  """Returns whether the float tensor contains all finite entries.

  Args:
    tensor: A float tensor. This cannot be an int tensor, or else
      tf.math.is_finite() will fail!
  """
  return bool(tf.reduce_all(tf.math.is_finite(tensor)))


def is_castable(to_cast, dtype):
  """Returns whether `to_cast` (a Value) can be safely casted to the dtype.

  This filtering strategy is a workaround for undefined behavior in TensorFlow
  (b/119633897).

  Args:
    to_cast: A Value object that would be casted.
    dtype: A Value containing a tf.DType that `to_cast` would be casted to.
  """
  if not dtype.is_int_dtype():
    return True  # We can always cast to a non-int dtype.

  to_cast_value = to_cast.value
  if to_cast.is_sparse_tensor:
    to_cast_value = to_cast.value.values

  if to_cast.is_tensor or to_cast.is_sparse_tensor:
    if not to_cast.has_float_dtype():
      return True  # Only float -> int is potentially unsafe.
    if not _check_tensor_finite(to_cast_value):
      return False  # Non-finite floats cannot be casted to int dtypes.
  elif to_cast.is_sequence:
    if to_cast.elem_type is float:
      if float('nan') in to_cast_value:
        return False  # inf and -inf will be caught by the min/max logic.
    elif to_cast.elem_type_is_tensor:
      return all(tf.size(element) and
                 is_castable(value_module.InputValue(element, 'dummy'), to_cast)
                 for element in to_cast_value)
    elif to_cast.elem_type_is_sparse_tensor:
      return all(tf.size(element.values) and
                 is_castable(value_module.InputValue(element, 'dummy'), to_cast)
                 for element in to_cast_value)
    else:
      return True  # Only lists of floats or float tensors can be unsafe.
  elif to_cast.type is float:
    if math.isnan(to_cast_value):
      return False
  else:
    return True

  min_int, max_int = tf_coder_utils.INT_DTYPE_MIN_MAX[dtype.value]

  # Floats are truncated when casted to int (nearest int in the zero direction).
  # Assuming min_int <= 0, the minimum safe float is (min_int - 1 + epsilon),
  # and the maximum safe float is (max_int + 1 - epsilon).
  return to_cast.min() > min_int - 1 and to_cast.max() < max_int + 1


def broadcastable(shape_1, shape_2):
  """Returns whether the two shapes are broadcastable."""
  return (not shape_1 or not shape_2 or
          all(x == y or x == 1 or y == 1
              for x, y in zip(shape_1[::-1], shape_2[::-1])))

# Constants for common filters. These are named with uppercase to reinforce the
# fact that these are constants and should be used as such, even though they are
# also technically functions.
# pylint: disable=invalid-name

# A filter that only keeps primitives.
PRIMITIVE_FILTER = operator.attrgetter('is_primitive')

# A filter that only keeps tf.DType objects.
DTYPE_FILTER = operator.attrgetter('is_dtype')

# A filter that only keeps sequences.
SEQUENCE_FILTER = operator.attrgetter('is_sequence')

# A filter that only keeps sequences of tensors.
TENSOR_SEQUENCE_FILTER = operator.attrgetter('elem_type_is_tensor')

# A filter that only keeps sequences of SparseTensors.
SPARSE_SEQUENCE_FILTER = operator.attrgetter('elem_type_is_sparse_tensor')

# A filter that only keeps tensors.
TENSOR_FILTER = operator.attrgetter('is_tensor')

# A filter that only keeps SparseTensors.
SPARSE_FILTER = operator.attrgetter('is_sparse_tensor')


def FLOAT_TENSOR_FILTER(arg_value):
  """Only keeps float tensors."""
  return arg_value.is_tensor and arg_value.has_float_dtype()


def NUMERIC_TENSOR_FILTER(arg_value):
  """Only keeps int and float tensors."""
  return arg_value.is_tensor and (arg_value.has_int_dtype() or
                                  arg_value.has_float_dtype())


def NONSCALAR_NUMERIC_TENSOR_FILTER(arg_value):
  """Only keeps non-scalar int and float tensors."""
  return NUMERIC_TENSOR_FILTER(arg_value) and len(arg_value.shape)


def INDICES_FILTER(arg_value):
  """Only keeps tensors/sequences containing ints suitable for indexing."""
  return ((arg_value.is_tensor and arg_value.has_int_dtype() or
           arg_value.elem_type is int) and
          arg_value.min() >= 0)


def AXIS_FILTER(arg_value):
  """Only keeps ints in the range [-1, limits.MAX_NUM_DIMENSIONS)."""
  return (arg_value.type is int and
          -1 <= arg_value.value < limits.MAX_NUM_DIMENSIONS)


def AXIS_SEQUENCE_FILTER(arg_value):
  """Only keeps sequences or 1D tensors of axis-like ints."""
  return (CONTAINS_INTS_1D_FILTER(arg_value) and
          len(arg_value.value) <= limits.MAX_NUM_DIMENSIONS and
          -1 <= arg_value.min() and
          arg_value.max() < limits.MAX_NUM_DIMENSIONS)


def PRIMITIVE_OR_SCALAR_TENSOR_FILTER(arg_value):
  """Only keeps primitives or scalar tensors."""
  return (arg_value.is_primitive or
          arg_value.is_tensor and arg_value.shape is None)


def NON_SCALAR_TENSOR_FILTER(arg_value):
  """Only keeps tensors that are not scalars."""
  return arg_value.is_tensor and arg_value.shape


def NOT_TENSOR_FILTER(arg_value):
  """Only keeps a value if it is not a Tensor or SparseTensor."""
  return not arg_value.is_tensor and not arg_value.is_sparse_tensor


def PRIMITIVE_OR_TENSOR_FILTER(arg_value):
  """Only keeps primitives and tensors."""
  return arg_value.is_primitive or arg_value.is_tensor


def TENSOR_1D_FILTER(arg_value):
  """Only keeps 1-D tensors."""
  return arg_value.is_tensor and len(arg_value.shape) == 1


def CONTAINS_INTS_FILTER(arg_value):
  """Only keeps int sequences or int tensors."""
  return arg_value.elem_type is int or arg_value.has_int_dtypes()


def CONTAINS_INTS_1D_FILTER(arg_value):
  """Only keeps int sequences or 1-D int tensors."""
  return (arg_value.elem_type is int or
          arg_value.has_int_dtype() and TENSOR_1D_FILTER(arg_value))


def TENSOR_LIKE_SEQUENCE_FILTER(arg_value):
  """Only keeps rectangular possibly-nested sequences of primitives."""
  return arg_value.is_sequence and arg_value.sequence_dtype is not None


def INT_OR_INT_TENSOR_FILTER(arg_value):
  """Only keeps int primitives or int tensors."""
  return (arg_value.type is int or
          (arg_value.is_tensor and not arg_value.shape and
           arg_value.has_int_dtype()))


def INT_LENGTH_FILTER(arg_value):
  """Only keeps int primitives or tensors representing a dimension length."""
  return (INT_OR_INT_TENSOR_FILTER(arg_value) and
          0 < int(arg_value.value) <= limits.MAX_DIMENSION_LENGTH)


def SHAPE_FILTER(arg_value):
  """Only keeps int sequences representing tensor shapes."""
  return (arg_value.elem_type is int and
          0 < len(arg_value.value) <= limits.MAX_NUM_DIMENSIONS and
          arg_value.min() > 0 and
          arg_value.max() <= limits.MAX_DIMENSION_LENGTH and
          arg_value.reduce_prod() <= limits.MAX_TENSOR_ELEMENTS)


def TENSOR_OR_SPARSE_FILTER(arg_value):
  """Only keeps Tensors and SparseTensors."""
  return arg_value.is_tensor or arg_value.is_sparse_tensor


def SQUARE_MATRIX_SIZE_FILTER(arg_value):
  """Ensures that an NxN matrix (N is the argument) is small enough."""
  if not INT_OR_INT_TENSOR_FILTER(arg_value):
    return False
  num_rows = int(arg_value.value)
  return (0 < num_rows <= limits.MAX_DIMENSION_LENGTH and
          num_rows ** 2 <= limits.MAX_TENSOR_ELEMENTS)


def SEQUENCE_MASK_LENGTHS_FILTER(arg_value):
  """The value must contain few ints with a small maximum."""
  # Only int tensors (not SparseTensors), or list of ints, are ok.
  if not (arg_value.is_tensor and arg_value.has_int_dtype()
          or arg_value.elem_type is int):
    return False
  max_value = arg_value.max()
  num_elements = arg_value.num_elements()
  return (num_elements > 0 and
          max_value * num_elements <= limits.MAX_TENSOR_ELEMENTS)


def PADDINGS_FILTER(arg_value):
  """Must be a [N, 2] shape int32 tensor or nested sequence of ints."""
  if arg_value.is_tensor:
    dtype = arg_value.dtype
    shape = arg_value.shape
  elif arg_value.is_sequence:
    dtype = arg_value.sequence_dtype
    shape = arg_value.sequence_shape
  else:
    return False
  if not (dtype == tf.int32 and
          shape and len(shape) == 2 and shape[1] == 2 and
          shape[0] <= limits.MAX_NUM_DIMENSIONS):
    return False
  return (0 <= arg_value.min() and
          arg_value.max() < limits.MAX_DIMENSION_LENGTH / 2)


def BATCH_DIMS_FILTER(arg_value):
  """Must be an int representing a number of batch dimensions."""
  return (arg_value.type is int and
          0 <= arg_value.value < limits.MAX_NUM_DIMENSIONS)


def SCATTER_INDICES_FILTER(arg_value):
  """Must be an int tensor appropriate for indices in scatter operations."""
  return (arg_value.is_tensor and
          arg_value.has_int_dtype and
          len(arg_value.shape) >= 2 and
          arg_value.shape[-1] <= limits.MAX_NUM_DIMENSIONS and
          arg_value.min() >= 0 and
          arg_value.max() < limits.MAX_DIMENSION_LENGTH)


def REPEATS_FILTER(arg_value):
  """Only keeps nonneg. int scalars/seqs. or 1-D int tensors with small sum."""
  return ((arg_value.type is int or CONTAINS_INTS_1D_FILTER(arg_value)) and
          arg_value.min() >= 0 and
          0 < arg_value.reduce_sum() <= limits.MAX_DIMENSION_LENGTH)


def SAME_DTYPES_APPLY_FILTER(arg_values):
  """Ensures that the first two arguments have the same dtype."""
  return arg_values[0].dtype == arg_values[1].dtype


def SAME_DTYPES_BROADCASTABLE_APPLY_FILTER(arg_values):
  """The two args must have the same dtypes and be broadcastable."""
  x, y = arg_values
  return x.dtype == y.dtype and broadcastable(x.shape, y.shape)


def SAME_SHAPES_APPLY_FILTER(arg_values):
  """Ensures that the first two arguments have the same shape."""
  return arg_values[0].shape == arg_values[1].shape


def TENSOR_AXIS_IN_RANGE_APPLY_FILTER(arg_values):
  """Ensures the axis is less than the rank of the tensor."""
  tensor, axis = arg_values
  return axis.value < len(tensor.shape)

# End of section for filter constants. pylint: enable=invalid-name


# LINT.IfChange(add_filters_to_function_operation)
def add_filters_to_function_operation(function_operation):
  """Adds filters to the FunctionOperation depending on its FilterGroup."""
  group = function_operation.function_info.filter_group

  if group == filter_group.FilterGroup.NONE:
    # Do nothing.
    pass

  elif group == filter_group.FilterGroup.SHAPE_1:
    function_operation.add_value_filters([SHAPE_FILTER])
  elif group == filter_group.FilterGroup.SEQUENCE_1:
    function_operation.add_value_filters([SEQUENCE_FILTER])
  elif group == filter_group.FilterGroup.TENSOR_1:
    function_operation.add_value_filters([TENSOR_FILTER])
  elif group == filter_group.FilterGroup.FLOATTENSOR_1:
    function_operation.add_value_filters([FLOAT_TENSOR_FILTER])
  elif group == filter_group.FilterGroup.NUMERICTENSOR_1:
    function_operation.add_value_filters([NUMERIC_TENSOR_FILTER])
  elif group == filter_group.FilterGroup.BOOLTENSOR_1:
    function_operation.add_value_filters([get_dtype_filter(tf.bool)])
  elif group == filter_group.FilterGroup.SPARSE_1:
    function_operation.add_value_filters([SPARSE_FILTER])
  elif group == filter_group.FilterGroup.NOT_TENSOR_1:
    function_operation.add_value_filters([NOT_TENSOR_FILTER])
  elif group == filter_group.FilterGroup.TENSOR_1D_1:
    function_operation.add_value_filters([TENSOR_1D_FILTER])
  elif group == filter_group.FilterGroup.PRIMITIVE_OR_TENSOR_1:
    function_operation.add_value_filters([PRIMITIVE_OR_TENSOR_FILTER])

  elif group == filter_group.FilterGroup.TENSOR_AXIS_2:
    function_operation.add_value_filters([TENSOR_FILTER, AXIS_FILTER])
    function_operation.set_apply_filter(TENSOR_AXIS_IN_RANGE_APPLY_FILTER)
  elif group == filter_group.FilterGroup.TENSOR_AXISSEQUENCE_2:
    function_operation.add_value_filters([TENSOR_FILTER, AXIS_SEQUENCE_FILTER])
    def _axis_sequence_in_range(arg_values):
      """The axes in the sequence must be in range given the tensor rank."""
      tensor, axes = arg_values
      return axes.max() < len(tensor.shape)
    function_operation.set_apply_filter(_axis_sequence_in_range)
  elif group == filter_group.FilterGroup.BOOLTENSOR_AXIS_2:
    function_operation.add_value_filters([get_dtype_filter(tf.bool),
                                          AXIS_FILTER])
    function_operation.set_apply_filter(TENSOR_AXIS_IN_RANGE_APPLY_FILTER)
  elif group == filter_group.FilterGroup.NUMERICTENSOR_AXIS_2:
    function_operation.add_value_filters([NUMERIC_TENSOR_FILTER, AXIS_FILTER])
    function_operation.set_apply_filter(TENSOR_AXIS_IN_RANGE_APPLY_FILTER)
  elif group == filter_group.FilterGroup.TENSORSEQUENCE_AXIS_2:
    function_operation.add_value_filters([TENSOR_SEQUENCE_FILTER,
                                          AXIS_FILTER])
  elif group == filter_group.FilterGroup.TENSOR_SHAPE_2:
    function_operation.add_value_filters([TENSOR_FILTER, SHAPE_FILTER])
  elif group == filter_group.FilterGroup.SHAPE_PRIMITIVE_2:
    function_operation.add_value_filters([SHAPE_FILTER, PRIMITIVE_FILTER])
  elif group == filter_group.FilterGroup.TENSOR_BOOLTENSOR_2:
    function_operation.add_value_filters([TENSOR_FILTER,
                                          get_dtype_filter(tf.bool)])
  elif group == filter_group.FilterGroup.CASTABLE_DTYPE_2:
    function_operation.add_value_filters([None, DTYPE_FILTER])
    function_operation.set_apply_filter(
        lambda arg_values: is_castable(arg_values[0], arg_values[1]))
  elif group == filter_group.FilterGroup.SAME_DTYPE_NUMERIC_BROADCASTABLE_2:
    function_operation.add_value_filters([NUMERIC_TENSOR_FILTER] * 2)
    function_operation.set_apply_filter(SAME_DTYPES_BROADCASTABLE_APPLY_FILTER)
  elif group == filter_group.FilterGroup.SAME_DTYPE_FLOAT_BROADCASTABLE_2:
    function_operation.add_value_filters([FLOAT_TENSOR_FILTER] * 2)
    function_operation.set_apply_filter(SAME_DTYPES_BROADCASTABLE_APPLY_FILTER)
  elif group == filter_group.FilterGroup.BOOLTENSOR_BROADCASTABLE_2:
    function_operation.add_value_filters([get_dtype_filter(tf.bool)] * 2)
    function_operation.set_apply_filter(SAME_DTYPES_BROADCASTABLE_APPLY_FILTER)
  elif group == filter_group.FilterGroup.SAME_SHAPE_ONE_SPARSE_2:
    def _same_shape_one_sparse_filter(arg_values):
      arg1, arg2 = arg_values
      return ((arg1.is_sparse_tensor or arg2.is_sparse_tensor) and
              SAME_SHAPES_APPLY_FILTER(arg_values))
    function_operation.add_value_filters([TENSOR_OR_SPARSE_FILTER] * 2)
    function_operation.set_apply_filter(_same_shape_one_sparse_filter)
  elif group == filter_group.FilterGroup.SAME_SHAPE_BOTH_SPARSE_2:
    function_operation.add_value_filters([SPARSE_FILTER] * 2)
    function_operation.set_apply_filter(SAME_SHAPES_APPLY_FILTER)
  elif group == filter_group.FilterGroup.AXIS_SPARSESEQUENCE_2:
    function_operation.add_value_filters([AXIS_FILTER, SPARSE_SEQUENCE_FILTER])
  elif group == filter_group.FilterGroup.SPARSE_AXIS_2:
    function_operation.add_value_filters([SPARSE_FILTER, AXIS_FILTER])
    function_operation.set_apply_filter(TENSOR_AXIS_IN_RANGE_APPLY_FILTER)
  elif group == filter_group.FilterGroup.SPARSE_SHAPE_2:
    function_operation.add_value_filters([SPARSE_FILTER, SHAPE_FILTER])
  elif group == filter_group.FilterGroup.SPARSE_PRIMITIVE_2:
    function_operation.add_value_filters([SPARSE_FILTER, PRIMITIVE_FILTER])
  elif group == filter_group.FilterGroup.SEGMENT_OPERATION_2:
    def _segment_ids_filter(arg_value):
      """Segment IDs must be a nonnegative nondecreasing 1D int tensor."""
      if not (arg_value.is_tensor and
              arg_value.has_int_dtype() and
              len(arg_value.shape) == 1 and
              arg_value.max() < limits.MAX_DIMENSION_LENGTH and
              arg_value.shape[0] > 0):
        return False
      elements = arg_value.numpy_tolist()
      return (0 <= elements[0] and
              all(x <= y for x, y in zip(elements, elements[1:])))
    function_operation.add_value_filters([NONSCALAR_NUMERIC_TENSOR_FILTER,
                                          _segment_ids_filter])
    def _segment_ids_right_length(arg_values):
      data, segment_ids = arg_values
      return (data.shape[0] == segment_ids.shape[0] and
              data.num_elements() / data.shape[0] * segment_ids.max()
              <= limits.MAX_TENSOR_ELEMENTS)
    function_operation.set_apply_filter(_segment_ids_right_length)

  elif group == filter_group.FilterGroup.SPARSE_INT_AXIS_3:
    function_operation.add_value_filters([SPARSE_FILTER,
                                          get_type_filter(int),
                                          AXIS_FILTER])
  elif group == filter_group.FilterGroup.SPARSE_AXIS_BOOL_3:
    function_operation.add_value_filters([SPARSE_FILTER,
                                          AXIS_FILTER,
                                          get_type_filter(bool)])
  elif group == filter_group.FilterGroup.UNSORTED_SEGMENT_OPERATION_3:
    def _segment_ids_filter(arg_value):
      """Must be a nonnegative nonscalar int tensor."""
      return (arg_value.is_tensor and
              arg_value.shape and arg_value.shape[0] and
              arg_value.has_int_dtype() and
              arg_value.min() >= 0 and
              arg_value.max() < limits.MAX_DIMENSION_LENGTH)
    function_operation.add_value_filters([NONSCALAR_NUMERIC_TENSOR_FILTER,
                                          _segment_ids_filter,
                                          INT_LENGTH_FILTER])
    def _unsorted_segment_apply_filter(arg_values):
      """Shapes must be compatible, num_segments must be large enough."""
      data, segment_ids, num_segments = arg_values
      return (data.shape[:len(segment_ids.shape)] == segment_ids.shape and
              segment_ids.max() < int(num_segments.value) and
              # Upper bound on the resulting tensor size.
              data.num_elements() * int(num_segments.value)
              <= limits.MAX_TENSOR_ELEMENTS)
    function_operation.set_apply_filter(_unsorted_segment_apply_filter)

  # Operations with other special handling.

  elif group == filter_group.FilterGroup.BINCOUNT_1:
    def _bincount_filter(arg_value):
      """The value must contain nonnegative ints with a small maximum."""
      # Must be an int tensor, lists of ints, or int primitive.
      if not (arg_value.is_tensor and arg_value.has_int_dtype() or
              arg_value.elem_type is int or
              arg_value.type is int):
        return False
      max_value = arg_value.max()
      min_value = arg_value.min()
      return min_value >= 0 and max_value <= limits.MAX_DIMENSION_LENGTH
    function_operation.add_value_filters([_bincount_filter])

  elif group == filter_group.FilterGroup.EYE_1:
    function_operation.add_value_filters([SQUARE_MATRIX_SIZE_FILTER])

  elif group == filter_group.FilterGroup.RANGE_1:
    function_operation.add_value_filters([INT_LENGTH_FILTER])

  elif group == filter_group.FilterGroup.SEQUENCE_MASK_1:
    function_operation.add_value_filters([SEQUENCE_MASK_LENGTHS_FILTER])

  elif group == filter_group.FilterGroup.SQUEEZE_1:
    def _squeezable_filter(arg_value):
      return TENSOR_FILTER(arg_value) and 1 in (arg_value.shape or [])
    function_operation.add_value_filters([_squeezable_filter])

  elif group == filter_group.FilterGroup.BROADCAST_TO_2:
    def _broadcast_to_apply_filter(arg_values):
      """Checks that the tensor is broadcastable to the shape."""
      tensor, shape = arg_values
      tensor_shape = tensor.shape
      shape_values = shape.value
      if not tensor_shape or len(tensor_shape) > len(shape_values):
        return False
      return broadcastable(tensor_shape, shape_values)
    function_operation.add_value_filters([TENSOR_FILTER, SHAPE_FILTER])
    function_operation.set_apply_filter(_broadcast_to_apply_filter)

  elif group == filter_group.FilterGroup.EXPAND_DIMS_2:
    function_operation.add_value_filters([TENSOR_FILTER, AXIS_FILTER])
    def _axis_in_range(arg_values):
      """Ensures the axis is at most the rank of the tensor."""
      tensor, axis = arg_values
      return axis.value <= len(tensor.shape)
    function_operation.set_apply_filter(_axis_in_range)

  elif group == filter_group.FilterGroup.EYE_ROWS_COLS_2:
    def _eye_rows_cols_apply_filter(arg_values):
      """Checks that the result will have a small number of elements."""
      num_rows, num_cols = arg_values
      return (int(num_rows.value) * int(num_cols.value)
              <= limits.MAX_TENSOR_ELEMENTS)
    function_operation.add_value_filters([INT_LENGTH_FILTER] * 2)
    function_operation.set_apply_filter(_eye_rows_cols_apply_filter)

  elif group == filter_group.FilterGroup.EYE_ROWS_DTYPE_2:
    function_operation.add_value_filters([SQUARE_MATRIX_SIZE_FILTER,
                                          DTYPE_FILTER])

  elif group == filter_group.FilterGroup.GATHER_2:
    function_operation.add_value_filters([NON_SCALAR_TENSOR_FILTER,
                                          INDICES_FILTER])
    def _indices_in_range(arg_values):
      params, indices = arg_values
      return (indices.max() < params.shape[0] and
              (params.num_elements() * indices.num_elements() / params.shape[0]
               <= limits.MAX_TENSOR_ELEMENTS))
    function_operation.set_apply_filter(_indices_in_range)

  elif group == filter_group.FilterGroup.GATHER_ND_2:
    function_operation.add_value_filters([NON_SCALAR_TENSOR_FILTER,
                                          INDICES_FILTER])
    def _gather_nd_2_apply_filter(arg_values):
      params, indices = arg_values
      indices_shape = (indices.shape if indices.is_tensor else
                       indices.sequence_shape)
      return (indices_shape and indices_shape[-1] <= len(params.shape) and
              indices.max() < max(params.shape) and
              # Upper bound on resulting tensor size.
              indices.num_elements() * params.num_elements()
              <= limits.MAX_TENSOR_ELEMENTS)
    function_operation.set_apply_filter(_gather_nd_2_apply_filter)

  elif group == filter_group.FilterGroup.MATMUL_2:
    def _numeric_min_rank_2_filter(arg_value):
      """Must be an int or float tensor of rank >= 2."""
      return arg_value.is_tensor and len(arg_value.shape) >= 2
    function_operation.add_value_filters([_numeric_min_rank_2_filter] * 2)
    function_operation.set_apply_filter(SAME_DTYPES_APPLY_FILTER)

  elif group == filter_group.FilterGroup.ONE_HOT_2:
    def _one_hot_indices_filter(arg_value):
      """Must contain ints and less than the max number of dimensions."""
      if arg_value.is_tensor:
        return (arg_value.has_int_dtype() and
                len(arg_value.shape) < limits.MAX_NUM_DIMENSIONS)
      else:
        return arg_value.type is int or arg_value.elem_type is int
    def _one_hot_apply_filter(arg_values):
      """Checks that the result will have a small number of elements."""
      indices, depth = arg_values
      return (indices.num_elements() * int(depth.value) <=
              limits.MAX_TENSOR_ELEMENTS)
    function_operation.add_value_filters([_one_hot_indices_filter,
                                          INT_LENGTH_FILTER])
    function_operation.set_apply_filter(_one_hot_apply_filter)

  elif group == filter_group.FilterGroup.PAD_2:
    function_operation.add_value_filters([TENSOR_FILTER, PADDINGS_FILTER])
    def _pad_2_apply_filter(arg_values):
      tensor, paddings = arg_values
      if paddings.is_tensor:
        paddings_shape = paddings.shape
      else:  # Implies paddings is a sequence.
        paddings_shape = paddings.sequence_shape
      return (tensor.shape and
              len(tensor.shape) == paddings_shape[0] and
              tf.reduce_prod(tf.add(tf.reduce_sum(paddings.value, axis=1),
                                    tensor.shape))
              <= limits.MAX_TENSOR_ELEMENTS)
    function_operation.set_apply_filter(_pad_2_apply_filter)

  elif group == filter_group.FilterGroup.REPEAT_2:
    function_operation.add_value_filters([PRIMITIVE_OR_TENSOR_FILTER,
                                          REPEATS_FILTER])
    def _repeat_2_apply_filter(arg_values):
      input_, repeats = arg_values
      input_num_elements = input_.num_elements()
      repeats_num_elements = repeats.num_elements()
      # Try to avoid creating too large tensors. The checks below should catch
      # egregious cases but aren't exhaustive.
      # If repeats has 1 element, then the length of the (flat) output equals
      # the number of repeats times the size of the input.
      return ((repeats_num_elements == 1 and
               repeats.max() * input_num_elements
               <= limits.MAX_DIMENSION_LENGTH) or
              # If repeats is a sequence, then the length of the output equals
              # the sum of the repeats, which is bounded in the value filter.
              repeats_num_elements == input_num_elements)
    function_operation.set_apply_filter(_repeat_2_apply_filter)

  elif group == filter_group.FilterGroup.SEARCHSORTED_2:
    def _sorted_last_dimension(arg_value):
      """Must be a numeric tensor that is sorted in the last dimension."""
      return (NONSCALAR_NUMERIC_TENSOR_FILTER(arg_value) and
              (arg_value.has_float_dtype() or
               arg_value.dtype in [tf.int32, tf.int64]) and
              bool(tf.reduce_all(tf.equal(arg_value.value,
                                          tf.sort(arg_value.value)))))
    function_operation.add_value_filters([_sorted_last_dimension,
                                          NONSCALAR_NUMERIC_TENSOR_FILTER])
    def _searchsorted_apply_filter(arg_values):
      """DTypes must match, dimension lengths equal except the last."""
      sorted_sequence, values = arg_values
      return (sorted_sequence.dtype == values.dtype and
              len(sorted_sequence.shape) == len(values.shape) and
              sorted_sequence.shape[:-1] == values.shape[:-1])
    function_operation.set_apply_filter(_searchsorted_apply_filter)

  elif group == filter_group.FilterGroup.SEQUENCE_MASK_2:
    function_operation.add_value_filters([SEQUENCE_MASK_LENGTHS_FILTER,
                                          INT_LENGTH_FILTER])
    def _sequence_mask_apply_filter(arg_values):
      """Checks that the result will have a small number of elements."""
      lengths, maxlen = arg_values
      return (lengths.num_elements() * int(maxlen.value)
              <= limits.MAX_TENSOR_ELEMENTS)
    function_operation.set_apply_filter(_sequence_mask_apply_filter)

  elif group == filter_group.FilterGroup.SPARSE_RETAIN_2:
    def _boolean_sequence_filter(arg_value):
      """Only keeps 1D boolean tensors, or boolean sequences."""
      if arg_value.is_tensor:
        return len(arg_value.shape) == 1 and arg_value.dtype == tf.bool
      return arg_value.elem_type == bool
    def _sparse_retain_apply_filter(arg_values):
      """Checks that to_retain is valid."""
      sp_input, to_retain = arg_values
      length = (
          to_retain.shape[0] if to_retain.is_tensor else len(to_retain.value))
      if to_retain.is_tensor:
        length = to_retain.shape[0]
      else:  # Must be a boolean sequence.
        length = len(to_retain.value)
      return length == int(sp_input.value.indices.shape[0])
    function_operation.add_value_filters([SPARSE_FILTER,
                                          _boolean_sequence_filter])
    function_operation.set_apply_filter(_sparse_retain_apply_filter)

  elif group == filter_group.FilterGroup.SPARSE_TO_INDICATOR_2:
    def _sparse_to_indicator_apply_filter(arg_values):
      """Checks that the result will be small."""
      sp_input, vocab_size = arg_values
      vocab_size_int = vocab_size.value
      if not sp_input.shape:
        return False
      if (vocab_size_int > limits.MAX_DIMENSION_LENGTH or
          vocab_size_int <= 0):
        return False
      output_size = vocab_size_int * functools.reduce(
          operator.mul, sp_input.shape[:-1], 1)
      return output_size <= limits.MAX_TENSOR_ELEMENTS
    function_operation.add_value_filters([SPARSE_FILTER, get_type_filter(int)])
    function_operation.set_apply_filter(_sparse_to_indicator_apply_filter)

  elif group == filter_group.FilterGroup.SPARSE_TRANSPOSE_2:
    def _sparse_transpose_apply_filter(arg_values):
      """Checks that perm has length equal to the number of a's dimensions."""
      a, perm = arg_values
      perm_len = perm.shape[0] if perm.is_tensor else len(perm.value)
      return perm_len == len(a.shape)
    function_operation.add_value_filters([SPARSE_FILTER,
                                          CONTAINS_INTS_1D_FILTER])
    function_operation.set_apply_filter(_sparse_transpose_apply_filter)

  elif group == filter_group.FilterGroup.SQUEEZE_2:
    def _very_squeezable_filter(arg_value):
      """Keeps tensors with more than 1 squeezable dimension."""
      # If a tensor only has 1 squeezable dimension, then this operation is
      # useless because it is simpler to use the one-arg version of squeeze.
      return TENSOR_FILTER(arg_value) and (arg_value.shape or []).count(1) >= 2
    function_operation.add_value_filters([_very_squeezable_filter,
                                          AXIS_FILTER])
    def _squeeze_2_apply_filter(arg_values):
      tensor, axis = arg_values
      return (axis.value < len(tensor.shape) and
              tensor.shape[axis.value] == 1)
    function_operation.set_apply_filter(_squeeze_2_apply_filter)

  elif group == filter_group.FilterGroup.TILE_2:
    def _tile_apply_filter(arg_values):
      """Checks that the result will have a small number of elements."""
      tensor, multiples = arg_values
      if multiples.is_sequence:
        dims = len(multiples.value)
      else:  # Tensor case.
        dims = multiples.shape[0]
      return (multiples.min() > 0 and
              multiples.max() > 1 and
              dims == len(tensor.shape) and
              multiples.reduce_prod() * tensor.num_elements()
              <= limits.MAX_TENSOR_ELEMENTS)
    function_operation.add_value_filters([TENSOR_FILTER,
                                          CONTAINS_INTS_1D_FILTER])
    function_operation.set_apply_filter(_tile_apply_filter)

  elif group == filter_group.FilterGroup.TOP_K_2:
    function_operation.add_value_filters([NONSCALAR_NUMERIC_TENSOR_FILTER,
                                          INT_LENGTH_FILTER])
    def _top_k_apply_filter(arg_values):
      """The tensor must have more than k items in the last dimension."""
      tensor, k = arg_values
      return int(k.value) < tensor.shape[-1]
    function_operation.set_apply_filter(_top_k_apply_filter)

  elif group == filter_group.FilterGroup.TRANSPOSE_2:
    def _transpose_apply_filter(arg_values):
      """Checks that perm has length equal to the number of a's dimensions."""
      a, perm = arg_values
      perm_len = perm.shape[0] if perm.is_tensor else len(perm.value)
      return perm_len == len(a.shape)
    function_operation.add_value_filters([TENSOR_FILTER,
                                          CONTAINS_INTS_1D_FILTER])
    function_operation.set_apply_filter(_transpose_apply_filter)

  elif group == filter_group.FilterGroup.CLIP_BY_VALUE_3:
    function_operation.add_value_filters([NUMERIC_TENSOR_FILTER,
                                          get_types_filter((int, float)),
                                          get_types_filter((int, float))])
    def _nondecreasing_clips(arg_values):
      _, min_clip, max_clip = arg_values
      return min_clip.value <= max_clip.value
    function_operation.set_apply_filter(_nondecreasing_clips)

  elif group == filter_group.FilterGroup.GATHER_ND_3:
    function_operation.add_value_filters([NON_SCALAR_TENSOR_FILTER,
                                          INDICES_FILTER,
                                          BATCH_DIMS_FILTER])
    def _gather_nd_3_apply_filter(arg_values):
      params, indices, batch_dims = arg_values
      batch_dims_int = batch_dims.value
      indices_shape = (indices.shape if indices.is_tensor else
                       indices.sequence_shape)
      return (
          batch_dims_int < min(len(indices_shape), len(params.shape)) and
          params.shape[:batch_dims_int] == indices_shape[:batch_dims_int] and
          indices_shape and
          indices_shape[-1] <= len(params.shape) - batch_dims_int and
          indices.max() < max(params.shape) and
          # Upper bound on resulting tensor size.
          indices.num_elements() * params.num_elements()
          <= limits.MAX_TENSOR_ELEMENTS)
    function_operation.set_apply_filter(_gather_nd_3_apply_filter)

  elif group == filter_group.FilterGroup.PAD_3:
    function_operation.add_value_filters([TENSOR_FILTER,
                                          PADDINGS_FILTER,
                                          PRIMITIVE_OR_SCALAR_TENSOR_FILTER])
    def _pad_3_apply_filter(arg_values):
      tensor, paddings, constant_values = arg_values
      if paddings.is_tensor:
        paddings_shape = paddings.shape
      else:  # Implies paddings is a sequence.
        paddings_shape = paddings.sequence_shape
      return (tensor.shape and len(tensor.shape) == paddings_shape[0] and
              (constant_values.is_primitive or
               constant_values.dtype == tensor.dtype) and
              tf.reduce_prod(tf.add(tf.reduce_sum(paddings.value, axis=1),
                                    tensor.shape))
              <= limits.MAX_TENSOR_ELEMENTS)
    function_operation.set_apply_filter(_pad_3_apply_filter)

  elif group == filter_group.FilterGroup.RANGE_3:
    def _range_3_apply_filter(arg_values):
      """Checks that the range will end up having a small number of elements."""
      start, limit, delta = arg_values
      return (delta.value != 0 and
              0 < len(range(start.value, limit.value, delta.value))
              <= limits.MAX_DIMENSION_LENGTH)
    function_operation.add_value_filters([get_type_filter(int)] * 3)
    function_operation.set_apply_filter(_range_3_apply_filter)

  elif group == filter_group.FilterGroup.ROLL_3:
    # The case where the shift and axis are both single integers.
    function_operation.add_value_filters([TENSOR_FILTER,
                                          INT_OR_INT_TENSOR_FILTER,
                                          AXIS_FILTER])
    # The case where the shift and axis are both sequences of integers.
    function_operation.add_value_filters([TENSOR_FILTER,
                                          CONTAINS_INTS_1D_FILTER,
                                          AXIS_SEQUENCE_FILTER])
    def _roll_apply_filter(arg_values):
      tensor, shift, axis = arg_values
      if axis.type is int:
        return axis.value < len(tensor.shape)
      else:
        return (len(axis.value) == len(shift.value) and
                axis.max() < len(tensor.shape))
    function_operation.set_apply_filter(_roll_apply_filter)

  elif group == filter_group.FilterGroup.REPEAT_3:
    function_operation.add_value_filters([NON_SCALAR_TENSOR_FILTER,
                                          REPEATS_FILTER,
                                          AXIS_FILTER])
    def _repeat_3_apply_filter(arg_values):
      input_, repeats, axis = arg_values
      repeats_num_elements = repeats.num_elements()
      # Try to avoid creating too large tensors. The checks below should catch
      # egregious cases but aren't exhaustive.
      return (axis.value < len(input_.shape) and
              # If repeats has 1 element, then the size of the output equals the
              # number of repeats times the size of the input.
              ((repeats_num_elements == 1 and
                repeats.max() * input_.num_elements()
                <= limits.MAX_TENSOR_ELEMENTS) or
               # If repeats is a sequence, then the length of the axis equals
               # the sum of the repeats, which is bounded in the value filter.
               repeats_num_elements == input_.shape[axis.value]))
    function_operation.set_apply_filter(_repeat_3_apply_filter)

  elif group == filter_group.FilterGroup.SCATTER_ND_3:
    function_operation.add_value_filters([SCATTER_INDICES_FILTER,
                                          TENSOR_FILTER,
                                          SHAPE_FILTER])
    def _scatter_nd_apply_filter(arg_values):
      indices, updates, shape = arg_values
      index_depth = indices.shape[-1]
      return (index_depth <= len(shape.value) and
              indices.max() < shape.max() and
              updates.shape == (indices.shape[:-1] +
                                list(shape.value[index_depth:])))
    function_operation.set_apply_filter(_scatter_nd_apply_filter)

  elif group == filter_group.FilterGroup.SPARSE_SLICE_3:
    def _contains_ints_1d_nonnegative_filter(arg_value):
      return CONTAINS_INTS_1D_FILTER(arg_value) and arg_value.min() >= 0
    def _sparse_slice_apply_filter(arg_values):
      sp_input, start, size = arg_values
      start_len = start.shape[0] if start.is_tensor else len(start.value)
      size_len = size.shape[0] if size.is_tensor else len(size.value)
      return (
          start_len == size_len == len(sp_input.shape)
          and all(int(start) < dim_length
                  for start, dim_length in zip(start.value, sp_input.shape)))
    function_operation.add_value_filters([SPARSE_FILTER,
                                          CONTAINS_INTS_1D_FILTER,
                                          _contains_ints_1d_nonnegative_filter])
    function_operation.set_apply_filter(_sparse_slice_apply_filter)

  elif group == filter_group.FilterGroup.SPARSETENSOR_3:
    def _indices_filter(arg_value):
      """Must be an int64 tensor with shape [num_values, num_dimensions]."""
      return (arg_value.is_tensor and
              arg_value.dtype == tf.int64 and
              len(arg_value.shape) == 2 and
              arg_value.shape[1] <= limits.MAX_NUM_DIMENSIONS)
    def _dense_shape_filter(arg_value):
      """A list of ints, or a 1D int64 tensor, representing a shape."""
      if not (arg_value.elem_type is int or
              (arg_value.is_tensor and arg_value.dtype == tf.int64 and
               len(arg_value.shape) == 1)):
        return False
      return (0 < arg_value.num_elements() <= limits.MAX_NUM_DIMENSIONS and
              arg_value.min() > 0 and
              arg_value.max() <= limits.MAX_DIMENSION_LENGTH)
    function_operation.add_value_filters([_indices_filter,
                                          TENSOR_1D_FILTER,
                                          _dense_shape_filter])
    def _sparsetensor_apply_filter(arg_values):
      """Shapes must be compatible."""
      indices, values, dense_shape = arg_values
      return (indices.shape[0] == values.shape[0] and
              indices.shape[1] == len(dense_shape.value))
    function_operation.set_apply_filter(_sparsetensor_apply_filter)

  elif group == filter_group.FilterGroup.TENSOR_SCATTER_ND_UPDATE_3:
    function_operation.add_value_filters([TENSOR_FILTER,
                                          SCATTER_INDICES_FILTER,
                                          TENSOR_FILTER])
    def _tensor_scatter_nd_update_apply_filter(arg_values):
      tensor, indices, updates = arg_values
      index_depth = indices.shape[-1]
      return (updates.dtype == tensor.dtype and
              index_depth <= len(tensor.shape) and
              indices.max() < max(tensor.shape) and
              updates.shape == (list(indices.shape[:-1]) +
                                tensor.shape[index_depth:]))
    function_operation.set_apply_filter(_tensor_scatter_nd_update_apply_filter)

  elif group == filter_group.FilterGroup.TENSORDOT_3:
    def _tensordot_arg_3_filter(arg_value):
      """The argument "axes" must have axis-like ints and the right shape."""
      if arg_value.type is int:
        # An int N means "sum over the last N axes of a and the first N axes of
        # b in order", so 0 <= N <= maximum rank.
        return 0 <= arg_value.value <= limits.MAX_NUM_DIMENSIONS
      if arg_value.elem_type is int:
        # List of length 2 is ok, elements must be valid axes.
        return (len(arg_value.value) == 2 and
                0 <= arg_value.min() and
                arg_value.max() < limits.MAX_NUM_DIMENSIONS)
      # Otherwise, must be an int tensor of shape [2] or [2, k].
      return (arg_value.is_tensor and
              arg_value.has_int_dtype() and
              1 <= len(arg_value.shape) <= 2 and
              arg_value.shape[0] == 2 and
              0 <= arg_value.min() and
              arg_value.max() < limits.MAX_NUM_DIMENSIONS)
    function_operation.add_value_filters([NONSCALAR_NUMERIC_TENSOR_FILTER,
                                          NONSCALAR_NUMERIC_TENSOR_FILTER,
                                          _tensordot_arg_3_filter])
    def _tensordot_apply_filter(arg_value):
      """First two tensors must have same dtype, and axes must be in range."""
      a, b, axes = arg_value
      if (a.dtype != b.dtype or
          # This check is overly conservative for the sake of efficiency; the
          # resulting number of elements is most likely smaller but will take
          # effort to compute more precisely.
          a.num_elements() * b.num_elements() > limits.MAX_TENSOR_ELEMENTS):
        return False
      a_rank = len(a.shape)
      b_rank = len(b.shape)
      min_rank = min(a_rank, b_rank)
      if axes.type is int:
        return axes.value <= min_rank
      elif axes.is_sequence or len(axes.shape) == 1:
        # axes is a list or tensor of shape [2].
        return axes.max() < min_rank
      else:  # axes is a tensor of shape [2, k].
        return (axes.shape[1] <= min_rank and
                tf_coder_utils.max_tensor_value(axes.value[0]) < a_rank and
                tf_coder_utils.max_tensor_value(axes.value[1]) < b_rank)

    function_operation.set_apply_filter(_tensordot_apply_filter)

  elif group == filter_group.FilterGroup.WHERE_3:
    def _where_apply_filter(arg_values):
      """Ensures that the last two arguments have matching shapes and dtypes."""
      _, x, y = arg_values
      return x.shape == y.shape and x.dtype == y.dtype
    function_operation.add_value_filters([get_dtype_filter(tf.bool),
                                          TENSOR_FILTER,
                                          TENSOR_FILTER])
    function_operation.set_apply_filter(_where_apply_filter)

  elif group == filter_group.FilterGroup.GATHER_4:
    function_operation.add_value_filters([NON_SCALAR_TENSOR_FILTER,
                                          INDICES_FILTER,
                                          AXIS_FILTER,
                                          BATCH_DIMS_FILTER])
    def _gather_4_apply_filter(arg_values):
      """Checks many constraints mentioned in the tf.gather() documentation."""
      params, indices, axis, batch_dims = arg_values
      axis_int = axis.value
      batch_dims_int = batch_dims.value
      indices_shape = (indices.shape if indices.is_tensor else
                       indices.sequence_shape)
      return (
          batch_dims_int < min(len(indices_shape), len(params.shape)) and
          (axis_int < 0 or axis_int >= batch_dims_int) and
          axis_int < len(params.shape) and
          # Upper bound on the size of the result.
          (params.num_elements() * indices.num_elements()
           / params.shape[axis_int] <= limits.MAX_TENSOR_ELEMENTS) and
          params.shape[:batch_dims_int] == indices_shape[:batch_dims_int] and
          indices.max() < params.shape[axis_int])
    function_operation.set_apply_filter(_gather_4_apply_filter)

  else:
    raise ValueError('Unknown filter group {} for FunctionOperation {}.'.format(
        group, function_operation.name))

# LINT.ThenChange()
# It is reasonable to strengthen or relax a filtering strategy here without
# involving a change to the filter groups.



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for function_operation.py."""

from absl import logging
from absl.testing import absltest
import tensorflow as tf
from tf_coder import filter_group
from tf_coder import tensor_limits as limits
from tf_coder import tf_functions
from tf_coder.value_search import function_operation
from tf_coder.value_search import value
from tf_coder.value_search import value_search_settings as settings_module


class FunctionOperationTest(absltest.TestCase):

  def setUp(self):
    super(FunctionOperationTest, self).setUp()
    self.settings = settings_module.default_settings()

  def test_metadata(self):
    operation = function_operation.FunctionOperation(
        tf_functions.FunctionInfo(name='tf.reduce_sum(input_tensor, axis)',
                                  filter_group=filter_group.FilterGroup.NONE,
                                  weight=2))
    docstring = operation.metadata.docstring
    self.assertIn('Computes the sum of elements across dimensions of a tensor.',
                  docstring)
    self.assertIn('tf.reduce_sum(input_tensor, axis)', docstring)
    self.assertIn('reduce sum', docstring)
    self.assertIn('input tensor', docstring)

  def test_apply_succeeds(self):
    operation = function_operation.FunctionOperation(
        tf_functions.FunctionInfo(name='tf.reduce_sum(input_tensor, axis)',
                                  filter_group=filter_group.FilterGroup.NONE,
                                  weight=2))
    input_tensor = value.ConstantValue(tf.constant([[1, 3, 4],
                                                    [50, 20, 80]]))
    axis_0 = value.ConstantValue(0)
    axis_1 = value.ConstantValue(1)
    self.assertEqual(operation.apply([input_tensor, axis_0], self.settings),
                     value.ConstantValue(tf.constant([51, 23, 84])))
    self.assertEqual(operation.apply([input_tensor, axis_1], self.settings),
                     value.ConstantValue(tf.constant([8, 150])))

  def test_apply_returns_none_if_exception(self):
    operation = function_operation.FunctionOperation(
        tf_functions.FunctionInfo(name='tf.reduce_sum(input_tensor, axis)',
                                  filter_group=filter_group.FilterGroup.NONE,
                                  weight=2))
    input_tensor = value.ConstantValue(tf.constant([[1, 3, 4],
                                                    [50, 20, 80]]))
    axis_2 = value.ConstantValue(2)
    self.assertIsNone(operation.apply([input_tensor, axis_2], self.settings))

  def test_apply_returns_none_if_bad_value(self):
    operation = function_operation.FunctionOperation(
        tf_functions.FunctionInfo(name='tf.one_hot(indices, depth)',
                                  filter_group=filter_group.FilterGroup.NONE,
                                  weight=2))
    indices = value.ConstantValue(
        tf.ones([limits.MAX_DIMENSION_LENGTH]))
    depth = value.ConstantValue(limits.MAX_TENSOR_ELEMENTS)
    self.assertIsNone(operation.apply([indices, depth], self.settings))

  def test_reconstruct_expression(self):
    operation = function_operation.FunctionOperation(
        tf_functions.FunctionInfo(name='tf.reduce_sum(input_tensor, axis)',
                                  filter_group=filter_group.FilterGroup.NONE,
                                  weight=2))
    arg_1 = value.InputValue([[1, 3], [50, 20]], 'my_input')
    arg_2 = value.ConstantValue('tf-coder')
    self.assertEqual(operation.reconstruct_expression([arg_1, arg_2]),
                     "tf.reduce_sum(my_input, axis='tf-coder')")

if __name__ == '__main__':
  logging.set_verbosity(logging.ERROR)

  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for tf_functions.py."""

from absl.testing import absltest
from absl.testing import parameterized
import funcsigs
from tf_coder import filter_group
from tf_coder import tf_coder_utils
from tf_coder import tf_functions

# FunctionInfo names should match this.
FUNCTION_INFO_NAME_REGEX = r'\w+(\.\w+)*\(\w+(, \w+(=[^,=()]+)?)*\)'


class TfFunctionsTest(parameterized.TestCase):

  def _check_function(self, function_name, usable_args, constant_kwargs):
    """Checks for errors in one entry of tf_functions.TF_FUNCTIONS."""
    try:
      func_obj = tf_coder_utils.get_tf_function(function_name)
    except ValueError:
      func_obj = None
    self.assertIsNotNone(func_obj,
                         'Could not find function {}.'.format(function_name))

    self.assertLen(set(usable_args), len(usable_args),
                   'Function {} has duplicate usable arguments.'.format(
                       function_name))

    parameters = funcsigs.signature(func_obj).parameters
    for param_name in parameters:
      param = parameters[param_name]
      if param.default is param.empty:
        self.assertIn(param_name, usable_args,
                      "Function {} is missing required argument '{}'.".format(
                          function_name, param_name))

    ordered_param_names = list(parameters)
    last_index = -1
    for arg_name in usable_args:
      self.assertIn(arg_name, ordered_param_names,
                    "Function {} has invalid argument '{}'.".format(
                        function_name, arg_name))
      cur_index = ordered_param_names.index(arg_name)
      self.assertGreater(cur_index, last_index,
                         "Function {} has argument '{}' out of order.".format(
                             function_name, arg_name))
      last_index = cur_index

    for kwarg_name in constant_kwargs:
      self.assertIn(kwarg_name, ordered_param_names,
                    "Function {} has invalid kwarg '{}'.".format(
                        function_name, kwarg_name))
      self.assertNotIn(kwarg_name, usable_args,
                       "Function {} has repeated argument '{}'.".format(
                           function_name, kwarg_name))

  def test_check_function_passes(self):
    self._check_function('tf.argsort', ['values', 'axis'],
                         {'direction': 'DESCENDING', 'stable': True})

  @parameterized.named_parameters(
      ('bad_function', 'tf.this_function_does_not_exist', ['x', 'y'], {}),
      ('duplicate_args', 'tf.add', ['x', 'y', 'x'], {}),
      ('missing_arg', 'tf.add', ['x'], {}),
      ('invalid_arg', 'tf.add', ['x', 'y', 'z'], {}),
      ('out_of_order', 'tf.add', ['y', 'x'], {}),
      ('duplicate_kwarg', 'tf.argsort', ['values', 'axis'], {'axis': -1}),
      ('invalid_kwarg', 'tf.argsort', ['values', 'axis'], {'invalid': True}))
  def test_check_function_fails(self, function_name, usable_args,
                                constant_kwargs):
    with self.assertRaises(AssertionError):
      self._check_function(function_name, usable_args, constant_kwargs)

  @parameterized.named_parameters(
      ('not_tf', 'np.foo(axis)'),
      ('nested_modules', 'tf.nn.foo(tensor, axis)'),
      ('no_module', 'foo(tensor, axis)'),
      ('string_literal', 'foo(tensor, axis, baz="a constant string")'),
      ('boolean_literal', 'foo(tensor, axis, baz=False)'),
      ('two_literals', 'foo(tensor, bar=[], baz=1.0)'))
  def test_function_info_name_regex_passes(self, good_name):
    self.assertRegex(good_name, FUNCTION_INFO_NAME_REGEX)

  @parameterized.named_parameters(
      ('bad_characters', 'tf.foo(axis=1)'),
      ('extra_spaces_1', 'tf.foo(tensor,  axis)'),
      ('extra_spaces_2', 'tf.foo( tensor, axis)'),
      ('extra_spaces_3', 'tf.foo (tensor, axis)'),
      ('missing_space', 'tf.foo(tensor,axis)'),
      ('missing_middle_arg', 'tf.foo(tensor, , axis)'),
      ('missing_last_arg', 'tf.foo(tensor, )'),
      ('no_args', 'tf.foo()'),
      ('no_parens', 'tf.foo'),
      ('empty_literal', 'tf.foo(a, x=)'),
      ('literal_with_bad_char', 'tf.foo(a, x=",")'))
  def test_function_info_name_regex_fails(self, bad_name):
    self.assertNotRegex(bad_name, FUNCTION_INFO_NAME_REGEX)

  @parameterized.named_parameters(
      ('tf_functions', tf_functions.TF_FUNCTIONS),
      ('sparse_functions', tf_functions.SPARSE_FUNCTIONS))
  def test_function_lists(self, function_list):
    for function_info in function_list:
      self.assertRegex(function_info.name, FUNCTION_INFO_NAME_REGEX)
      self.assertIsInstance(function_info.filter_group,
                            filter_group.FilterGroup)
      self.assertIsInstance(function_info.weight, int)

      function_name, usable_args, constant_kwargs = (
          tf_functions.parse_function_info_name(function_info))
      self._check_function(function_name, usable_args, constant_kwargs)

  def test_parse_function_info_name(self):
    function_info = tf_functions.FunctionInfo(
        name='tf.foo.bar(tensor, axis, baz=True)',
        filter_group=filter_group.FilterGroup.NONE,
        weight=1)
    self.assertEqual(tf_functions.parse_function_info_name(function_info),
                     ('tf.foo.bar', ['tensor', 'axis'], {'baz': True}))

  @parameterized.named_parameters(
      ('no_open_paren', 'tf.foo.bar tensor, axis)'),
      ('multiple_open_parens', 'tf.foo.bar((tensor, axis)'),
      ('no_close_paren', 'tf.foo.bar(tensor, axis'),
      ('close_paren_not_at_end', 'tf.foo.bar(tensor, axis) '))
  def test_parse_function_info_name_fails_for_bad_name(self, bad_name):
    function_info = tf_functions.FunctionInfo(
        name=bad_name,
        filter_group=filter_group.FilterGroup.NONE,
        weight=1)
    with self.assertRaises(ValueError):
      tf_functions.parse_function_info_name(function_info)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Config for the tensor features model."""

import tensorflow as tf
from tf_coder import tf_coder_utils
from tf_coder.datasets import collect_tensor_data
from tf_coder.value_search import all_operations


def get_config():
  """Gets the default config for the tensor features model."""
  cfg = {}

  # Model size.
  cfg['feedforward_size'] = 2048
  cfg['num_feedforward_layers'] = 2

  # Training hyperparams.
  cfg['batch_size'] = 1024
  cfg['learning_rate'] = 1e-5  # 1e-5 is better, 1e-4 is faster for debugging.
  cfg['global_norm_clip'] = 5
  cfg['num_epochs'] = 3

  # Loss function and options.

  # sigmoid_ce: Average the sigmoid cross entropy loss over all operations.
  # f_beta: Use a differentiable version of F_beta score as the loss.
  cfg['loss'] = 'f_beta'
  # If using f_beta loss, this is the value of beta. Because TF-Coder cares more
  # about recall than precision, this value should be >= 1.
  cfg['beta'] = 2
  # Whether to use weights to "balance" the prevalence of different ops. Let
  # `c_i` be the number of training examples operation `op_i` appears in. If
  # this option is True, then if `op_i` is used in a given example, we treat it
  # as actually appearing ((mean_j c_j) / c_i) times. This effectively upweights
  # rare operations and downweights common operations, such that each operation
  # now appears an equal (mean_j c_j) times across the dataset. Furthermore,
  # the total number of operations used across the dataset remains the same,
  # namely (sum_j c_j).
  cfg['weighted_ops'] = True
  # If 'mean', use (mean_j c_j) as the weight numerator. If 'max', use
  # (max_j c_j) instead, so that no operations are downweighted.
  cfg['weight_numerator'] = 'mean'
  # If using weights to balance the occurrences of ops, this is the maximum
  # weight. Weights that are too large may increase training instability.
  # However, we do have gradient clipping in place, so this may be unnecessary.
  cfg['max_weight'] = 10000

  # Details about the training data.
  operations = all_operations.get_operations(include_sparse_operations=True)
  cfg['num_ops'] = len(operations)
  cfg['num_kinds'] = 5
  cfg['num_dtypes'] = len(tf_coder_utils.INT_DTYPES +
                          tf_coder_utils.FLOAT_DTYPES + (tf.bool,))
  cfg['max_rank'] = collect_tensor_data.MAX_RANK
  cfg['num_shape_buckets'] = len(collect_tensor_data.COUNT_BOUNDARIES)
  cfg['num_float_buckets'] = len(collect_tensor_data.FLOAT_BOUNDARIES)
  cfg['num_count_buckets'] = len(collect_tensor_data.COUNT_BOUNDARIES)
  cfg['max_num_inputs'] = 3
  # Store the names of all operations used at the time of model training, so we
  # know how to interpret the model's predictions at test time, even if new
  # operations were added in between.
  cfg['operation_names'] = [op.name for op in operations]
  cfg['op_counts'] = [
      261791, 3219730, 49984, 468220, 70588, 671391, 304105, 76203, 361,
      12293742, 223803, 208334, 4469035, 0, 5228323, 395630, 514399, 8810328,
      175664, 982, 48877, 24769, 1429728, 15448, 124548, 675, 788727, 180186,
      770114, 198687, 7687, 26592, 773463, 443414, 26209, 183343, 397296,
      580331, 257051, 22749, 24187, 1229, 1727, 8011, 28719, 316277, 229433,
      54625, 5881, 23486, 2548, 33794, 466112, 1906853, 590050, 2780775, 29209,
      2393112, 4041, 342637, 39436, 9667, 1308, 2748, 752650, 17687, 5663,
      1246906, 1327261, 486676, 421095, 126426, 184234, 134355, 2565768,
      2124113, 5845, 49484, 338106, 63409, 6227, 9410, 953093, 124307, 878241,
      126966, 211238, 96981, 185479, 4384591, 800148, 119813, 188229, 5031788,
      1761827, 244750, 2493965, 2, 339585, 1145115, 4293436, 149049, 1058,
      342232, 687, 33192, 451, 31190, 1145201, 2621, 1642, 128519, 142314,
      22508, 16, 3660, 499, 48583, 359399, 118306, 3940, 9318, 3614, 543957,
      809717, 9070582, 2120948, 273953, 719889, 603656, 170092, 517624, 481419,
      2997663
  ]

  # Evaluation settings.
  cfg['eval_batch_size'] = 100000  # This is actually the entire eval set.
  cfg['eval_step_frequency'] = 100  # A multiple of summary_step_frequency.

  # Saving summaries and checkpoints.
  cfg['summary_step_frequency'] = 100
  cfg['save_step_frequency'] = 100
  cfg['keep_checkpoint_every_n_hours'] = 0.5
  cfg['allow_restore'] = False
  cfg['log_device_placement'] = False

  return cfg



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Benchmarks from real questions and tasks encountered by Googlers."""

# Avoid wrapping URLs and target programs to ease clicking and copying.
# pylint: disable=line-too-long

# Every function in this module takes no arguments and creates a benchmark.
# pylint: disable=missing-docstring

import tensorflow as tf
from tf_coder.benchmarks import benchmark


def google_01():
  examples = [
      benchmark.Example(
          inputs=[
              [0, 0, 0, 1, 3, 3],
          ],
          output=[[0, 0], [0, 1], [0, 2], [1, 0], [3, 0], [3, 1]]
      ),
  ]
  constants = []
  description = 'Convert index tensor into pairs for SparseTensor indexing'
  target_program = 'tf.cast(tf.where(tf.sequence_mask(tf.math.bincount(in1))), tf.int32)'
  source = 'From an internal Google chat room, 09/07/2018'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_01')


def google_02():
  examples = [
      benchmark.Example(
          inputs=[
              [[0.0, 1.0, 0.0, 0.0],
               [0.0, 1.0, 1.0, 0.0],
               [1.0, 1.0, 1.0, 1.0]],
          ],
          output=[[0.0, 1.0, 0.0, 0.0],
                  [0.0, 0.5, 0.5, 0.0],
                  [0.25, 0.25, 0.25, 0.25]]
      ),
  ]
  constants = []
  description = 'Divide each row by the sum of that row'
  target_program = 'tf.divide(in1, tf.expand_dims(tf.reduce_sum(in1, axis=1), 1))'
  source = 'Real task encountered by Googler, 11/01/2018'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_02')


def google_03():
  examples = [
      benchmark.Example(
          inputs=[
              tf.SparseTensor(
                  indices=[[0, 0, 0], [0, 1, 1], [1, 1, 1], [1, 1, 2]],
                  values=[1., 1., 1., 1.],
                  dense_shape=[2, 2, 800]),
          ],
          output=tf.SparseTensor(
              indices=[[0, 0, 0], [0, 1, 1]],
              values=[1., 1.],
              dense_shape=[1, 2, 800])
      ),
  ]
  constants = []
  description = 'Slice the first dimension of a SparseTensor'

  # Original approach, which can't be written as one expression.
  """
  reduced_sp_tensor = tf.sparse_retain(in1, tf.equal(in1.indices[:, 0], 0))
  output = tf.SparseTensor(indices=reduced_sp_tensor.indices,
                           values=reduced_sp_tensor.values,
                           dense_shape=tf.concat([tf.ones(1, dtype=tf.int64),
                                                  in1.dense_shape[1:]], 0))
  """  # pylint: disable=pointless-string-statement

  target_program = 'tf.sparse.slice(in1, tf.zeros(3, dtype=tf.int64), tf.concat([[1], in1.dense_shape[1:]], 0))'
  source = 'Real task encountered by Googler, 10/16/2018'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_03')


def google_04():
  examples = [
      benchmark.Example(
          inputs=[
              [111, 112, 121, 122, 131, 132, 211, 212, 221, 222, 231, 232,
               311, 312, 321, 322, 331, 332, 411, 412, 421, 422, 431, 432],
          ],
          output=[[[111, 112], [121, 122], [131, 132]],
                  [[211, 212], [221, 222], [231, 232]],
                  [[311, 312], [321, 322], [331, 332]],
                  [[411, 412], [421, 422], [431, 432]]]
      ),
  ]
  constants = []
  description = 'Reshape a flat array into a rank 3 tensor'
  target_program = 'tf.reshape(in1, shape=(4, 3, 2))'
  source = 'Real task encountered by Googler, 3/21/2019'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_04')


def google_05():
  examples = [
      benchmark.Example(
          inputs=[
              [[1, 2, 3, 4],
               [5, 6, 7, 8]],
          ],
          output=[[[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4],
                   [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]],
                  [[5, 6, 7, 8], [5, 6, 7, 8], [5, 6, 7, 8],
                   [5, 6, 7, 8], [5, 6, 7, 8], [5, 6, 7, 8]]]
      ),
  ]
  constants = [6]
  description = 'Repeat each input entry 6 times'
  target_program = 'tf.tile(tf.expand_dims(in1, 1), (1, 6, 1))'
  source = 'Real task encountered by Googler, 3/22/2019'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_05')


def google_06():
  examples = [
      benchmark.Example(
          inputs=[
              [0, 1, 1, 2, 3, 3, 3, 3],
              [1, 3, 4, 5, 10, 8, 9, 4],
          ],
          output=[1, 4, 5, 10]
      ),
  ]
  constants = []
  description = 'Take the max from each group of elements'
  target_program = 'tf.math.segment_max(in2, in1)'
  source = 'Real task encountered by Googler, 3/28/2019'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_06')


def google_07():
  examples = [
      benchmark.Example(
          inputs=[
              [0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 4, 4, 5, 5],
              [4, 1, 8, 2, 5, 7, 9, 3, 7, 3, 1, 42, 1, 2, 4, 0],
          ],
          output=[2, 3, 1, 0, 1, 0]
      ),
  ]
  constants = []
  description = 'Take the argmax of each group of elements'
  target_program = 'tf.cast(tf.argmax((tf.sequence_mask(tf.cumsum(tf.math.bincount(in1)), dtype=tf.int32) - tf.sequence_mask(tf.cumsum(tf.math.bincount(in1), exclusive=True), dtype=tf.int32, maxlen=in1.shape[0])) * in2, axis=1), tf.int32) - tf.cumsum(tf.math.bincount(in1), exclusive=True)'
  # Original question asker's solution below.
  """
  segment_lens = tf.math.bincount(in1)
  segment_ends = tf.scan(lambda a, x: a + x, segment_lens)
  segment_infos = tf.cast(tf.transpose(tf.stack([segment_lens, segment_ends])), dtype=tf.int64)
  def segment_argmax_fn(r):
    return tf.argmax(in2[r[1] - r[0]:r[1]])
  result = tf.cast(tf.map_fn(segment_argmax_fn, segment_infos), dtype=tf.int32)
  """  # pylint: disable=pointless-string-statement
  # Another Googler's solution below. Only works if in2 is positive.
  """
  segment_lens = tf.math.bincount(in1)
  cumulative_inclusive = tf.cumsum(segment_lens)
  cumulative_exclusive = tf.cumsum(segment_lens, exclusive=True)
  mask_inclusive = tf.sequence_mask(cumulative_inclusive, dtype=tf.int32)
  mask_exclusive = tf.sequence_mask(cumulative_exclusive, dtype=tf.int32, maxlen=in1.shape[0])
  masked_data = (mask_inclusive - mask_exclusive) * in2
  output = tf.cast(tf.argmax(masked_data, axis=1), tf.int32) - cumulative_exclusive
  """  # pylint: disable=pointless-string-statement
  source = 'Real task encountered by Googler, 3/29/2019'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_07')


def google_08():
  examples = [
      benchmark.Example(
          inputs=[
              [3, 4, 2, 1],
          ],
          output=[[1, 1, 1, 0, 0],
                  [1, 1, 1, 1, 0],
                  [1, 1, 0, 0, 0],
                  [1, 0, 0, 0, 0]]
      ),
  ]
  constants = [5]
  description = 'create a mask for sequences of the given lengths'
  target_program = 'tf.cast(tf.greater(tf.expand_dims(in1, 1), tf.range(5)), tf.int32)'
  source = 'Real task encountered by Googler, 4/11/2019'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_08')


def google_09():
  examples = [
      benchmark.Example(
          inputs=[
              [1, 1, 1, 0, 0, 2],
              [10, 20, 30, 14, 15, 26],
          ],
          output=[14, 15, 10, 20, 30, 26]
      ),
  ]
  constants = []
  description = 'sort the segments'
  target_program = 'tf.gather(in2, tf.argsort(in1, stable=True))'
  source = 'Real task encountered by Googler, 8/9/2019'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_09')


def google_10():
  examples = [
      benchmark.Example(
          inputs=[
              [10, 20, 0, 40, 0, 30],
              [1, 1, 0, 1, 0, 1],
          ],
          output=[10, 20, 40, 30]
      ),
  ]
  constants = []
  description = 'gather the marked elements'
  target_program = 'tf.boolean_mask(in1, tf.cast(in2, tf.bool))'
  source = ('Proposed by Googler at an internal demo on 8/13/2019, '
            'simplified slightly')
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_10')


def google_11():
  examples = [
      benchmark.Example(
          inputs=[
              [[1.0, 0.3, -4.2, 0.0, 2.1, 0.4],
               [-0.1, 0.0, 1.4, -1.0, 0.4, 0.0],
               [0.1, 0.0, 0.7, -0.3, 0.5, -0.1],
               [1.4, 2.5, 0.3, 0.01, 0.0, 1.2]],
          ],
          output=[4, 2, 3, 5]
      ),
  ]
  constants = []
  description = 'count the number of elements greater than 0 in each row'
  target_program = 'tf.reduce_sum(tf.cast(tf.greater(in1, 0), tf.int32), axis=1)'
  source = 'Real task encountered by Googler, 8/26/2019'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_11')


def google_12():
  examples = [
      benchmark.Example(
          inputs=[
              [[1.0, 0.3, -4.2, 0.0, 2.1],
               [-0.1, 0.0, 1.4, -1.0, 0.4],
               [0.1, 0.0, 0.7, -0.3, 0.5],
               [1.4, 2.5, 0.3, -0.1, 0.0]],
          ],
          output=[[1, 1, 0, 1, 0],
                  [0, 1, 0, 0, 1],
                  [1, 1, 1, 0, 1],
                  [0, 0, 1, 0, 1]]
      ),
  ]
  constants = []
  description = 'identify elements between 0 and 1'
  target_program = 'tf.cast(tf.logical_and(0 <= in1, in1 <= 1), tf.int32)'
  source = 'Real task encountered by Googler, 8/26/2019'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_12')


def google_13():
  examples = [
      benchmark.Example(
          inputs=[
              [[1, 2], [10, 20]],
              [[3, 4, 5], [30, 40, 50]],
          ],
          output=[[1, 2, 3, 4, 5], [10, 20, 30, 40, 50]]
      ),
  ]
  constants = []
  description = 'Concatenate batches of sequences'
  target_program = 'tf.concat([in1, in2], axis=1)'
  source = 'Real task encountered by Googler, 9/13/2019'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_13')


def google_14():
  examples = [
      benchmark.Example(
          inputs=[
              [[1, 3, 2, 0, 0], [4, 6, 5, 0, 0], [8, 7, 9, 0, 0]],
          ],
          output=[[0, 1, 3, 2, 0], [0, 4, 6, 5, 0], [0, 8, 7, 9, 0]]
      ),
  ]
  constants = []
  description = 'circular buffer'
  target_program = 'tf.roll(in1, 1, 1)'
  source = 'From internal Colab, 10/17/2019'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_14')


def google_15():
  examples = [
      benchmark.Example(
          inputs=[
              [[1, 3, 5, 7], [2, 4, 6, 8]],
          ],
          output=[[1, 3, 5, 7, 0], [2, 4, 6, 8, 0]]
      ),
  ]
  constants = []
  description = 'pad a zero column'
  target_program = "tf.pad(in1, [[0, 0], [0, 1]], 'CONSTANT')"
  source = 'Real task encountered by Googler, 10/23/2019'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_15')


def google_16():
  examples = [
      benchmark.Example(
          inputs=[
              [1, 2, 0, 3],
              [2, 1, 2, 3],
          ],
          output=[1, 1, 2, 0, 0, 3, 3, 3]
      ),
  ]
  constants = []
  description = 'replicate elements a given number of times'
  target_program = 'tf.gather(in1, tf.where(tf.sequence_mask(in2))[:, 0])'
  source = 'From an internal Google forum'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_16')


def google_17():
  examples = [
      benchmark.Example(
          inputs=[
              [True, False, False, True, False],
              [1, 2, 3, 4, 5],
          ],
          output=[1, -20, -30, 4, -50]
      ),
  ]
  constants = [-10]
  description = 'use bool tensor as condition'
  target_program = 'tf.where(in1, in2, tf.multiply(in2, -10))'
  source = 'From an internal Google forum'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_17')


def google_18():
  examples = [
      benchmark.Example(
          inputs=[
              [5, 7, -12, 10, 20],
              [1, 2, 3, 1, 2],
          ],
          output=[15, 27, -12, 15, 27]
      ),
  ]
  constants = []
  description = ('sum of elements in the first tensor but partitioned by the '
                 'second tensor')
  target_program = 'tf.linalg.matvec(tf.cast(tf.equal(in2[:, None], in2), tf.int32), in1)'
  source = 'From an internal Google forum'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_18')


def google_19():
  examples = [
      benchmark.Example(
          inputs=[
              [[11, 12, 13],
               [30, 20, 10],
               [77, 88, 99]],
              [[2, 0, 1],
               [1, 0, 2],
               [0, 2, 1]],
          ],
          output=[[12, 13, 11],
                  [20, 30, 10],
                  [77, 99, 88]]
      ),
  ]
  constants = []
  description = 'scatter a 2-D tensor with indices'
  target_program = 'tf.gather(in1, tf.argsort(in2, axis=1), batch_dims=1)'
  source = 'From an internal Google forum'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_19')


def google_20():
  examples = [
      benchmark.Example(
          inputs=[
              [10, 7, 4, 3, 2, 8],
          ],
          output=[5, 3, 2, 1, 0, 4]
      ),
  ]
  constants = []
  description = 'sort a tensor and return sorted index in original order'
  target_program = 'tf.cast(tf.argsort(tf.argsort(in1)), tf.int32)'
  source = 'From an internal Google forum'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_20')


def google_21():
  examples = [
      benchmark.Example(
          inputs={
              'tensor': [[1, 2, 3, 4, 5], [4, 5, 6, 7, 8], [7, 8, 9, 10, 11]],
              'indices': [[0, 0], [0, 2], [1, 1], [1, 3], [2, 2], [2, 4]],
              'updates': [[0, -2], [-1, -3], [-2, -4]],
          },
          output=[[0, 2, -2, 4, 5], [4, -1, 6, -3, 8], [7, 8, -2, 10, -4]]
      ),
  ]
  constants = []
  description = 'update a tensor at the given indices'
  target_program = 'tf.tensor_scatter_nd_update(tensor, indices, tf.reshape(updates, (-1,)))'
  source = 'Real task encountered by Googler, 12/15/2020'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_21')


def google_22():
  examples = [
      benchmark.Example(
          inputs=[
              [[0, 2], [1, 3], [2, 4]],
          ],
          output=[[0, 0], [0, 2], [1, 1], [1, 3], [2, 2], [2, 4]]
      ),
  ]
  constants = []
  description = 'pair with row index'
  target_program = 'tf.cast(tf.where(tf.reduce_max(tf.one_hot(in1, tf.reduce_max(in1) + 1), axis=1)), tf.int32)'
  source = 'Real task encountered by Googler, 12/15/2020'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_22')


# A template for easy copy/pasting. Copying an existing benchmark and replacing
# parts of it will lead to a state where the benchmark is half-correct, but not
# obviously so. Copy this template instead when creating new benchmarks.
"""

def google_NUMBER():
  examples = [
      benchmark.Example(
          inputs=[
              INPUT_1,
              INPUT_2,
          ],
          output=OUTPUT
      ),
  ]
  constants = [CONSTANTS]
  description = 'DESCRIPTION'
  target_program = 'SOLUTION_PROGRAM'
  source = 'PROBLEM_SOURCE'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='google_NUMBER')

"""  # pylint: disable=pointless-string-statement



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for tf_coder_utils.py."""

import collections

from absl import logging
from absl.testing import absltest
from absl.testing import parameterized
import tensorflow as tf
from tf_coder import tf_coder_utils


class TfCoderUtilsTest(parameterized.TestCase):

  def test_int_dtype_min_max_has_correct_keys(self):
    self.assertEqual(
        set(tf_coder_utils.INT_DTYPES), set(tf_coder_utils.INT_DTYPE_MIN_MAX))

  @parameterized.named_parameters(
      ('correct_usage_add', 'tf.add', tf.add),
      ('correct_usage_matmul', 'tf.matmul', tf.matmul),
      ('nested_modules', 'tf.nn.softmax', tf.nn.softmax))
  def test_get_tf_function(self, function_name, expected_result):
    self.assertEqual(
        tf_coder_utils.get_tf_function(function_name), expected_result)

  @parameterized.named_parameters(
      ('not_tf', 'np.add'),
      ('no_module', 'add'),
      ('bad_nested_module', 'tf.bad.softmax'),
      ('bad_nested_module_function', 'tf.nn.bad'),
      ('not_function', 'tf.this_is_not_a_function'))
  def test_get_tf_function_raises(self, function_name):
    with self.assertRaises(ValueError):
      tf_coder_utils.get_tf_function(function_name)

  def test_convert_to_tensor_for_tensor(self):
    original_tensor = tf.constant([1, 2, 3])
    tensor = tf_coder_utils.convert_to_tensor(original_tensor)
    self.assertIsInstance(tensor, tf.Tensor)
    self.assertEqual(original_tensor.numpy().tolist(), tensor.numpy().tolist())

  def test_convert_to_tensor_for_list(self):
    list_2d = [[1, 2, 3], [44, 55, 66]]
    tensor = tf_coder_utils.convert_to_tensor(list_2d)
    self.assertIsInstance(tensor, tf.Tensor)
    self.assertEqual(list_2d, tensor.numpy().tolist())

  def test_convert_to_tensor_for_scalar(self):
    scalar = 1.23
    tensor = tf_coder_utils.convert_to_tensor(scalar)
    self.assertIsInstance(tensor, tf.Tensor)
    self.assertAlmostEqual(scalar, tensor.numpy().tolist())

  @parameterized.named_parameters(
      ('1_dimension', (3), 3),
      ('2_dimensions', (3, 4), 12),
      ('3_dimensions', (3, 4, 5), 60))
  def test_num_tensor_elements_using_shape(self, shape, expected_result):
    self.assertEqual(
        tf_coder_utils.num_tensor_elements(tf.ones(shape)), expected_result)

  @parameterized.named_parameters(
      ('empty_list', [], 0),
      ('filled_list', [11, 12, 13], 3),
      ('scalar', 123, 1))
  def test_num_tensor_elements_using_content(self, content, expected_result):
    self.assertEqual(
        tf_coder_utils.num_tensor_elements(tf.constant(content)),
        expected_result)

  @parameterized.named_parameters(
      ('empty', [], float('-inf')),
      ('single_element_list', [12.5], 12.5),
      ('2_dimensions', [[-100, -50], [-123, -45]], -45.0),
      ('scalar', -12.5, -12.5))
  def test_max_tensor_value(self, content, expected_result):
    self.assertAlmostEqual(
        tf_coder_utils.max_tensor_value(tf.constant(content)), expected_result)

  @parameterized.named_parameters(
      ('empty', [], float('inf')),
      ('single_element_list', [12.5], 12.5),
      ('2_dimensions', [[-100, -50], [-123, -45]], -123.0),
      ('scalar', -12.5, -12.5))
  def test_min_tensor_value(self, content, expected_result):
    self.assertAlmostEqual(
        tf_coder_utils.min_tensor_value(tf.constant(content)), expected_result)

  def test_tensor_to_string(self):
    tensor = tf.constant([[1, 2], [3, 4]])
    self.assertEqual(
        tf_coder_utils.tensor_to_string(tensor), 'tf.int32:[[1, 2], [3, 4]]')

  def test_tensor_to_string_ugly_floats(self):
    # We do not know how to print floating-point numpy arrays in a prettier way.
    tensor = tf.constant([0.4, 0.7])
    self.assertEqual(
        tf_coder_utils.tensor_to_string(tensor),
        'tf.float32:[0.4000000059604645, 0.699999988079071]')

  @parameterized.named_parameters(
      ('zero_places', 0.9, 0, 'tf.float32:[1.0]'),
      ('round_down', 1.549, 1, 'tf.float32:[1.5]'),
      ('round_up', 1.249, 2, 'tf.float32:[1.25]'),
      ('not_rounded', 1.03125, 6, 'tf.float32:[1.03125]'))
  def test_tensor_to_string_performs_rounding(self, float_value, decimals,
                                              expected_result):
    tensor = tf.constant([float_value])
    self.assertEqual(
        tf_coder_utils.tensor_to_string(tensor, decimals=decimals),
        expected_result)

  def test_object_to_string_tensor(self):
    tensor = tf.constant([[1, 2], [3, 4]])
    self.assertEqual(
        tf_coder_utils.object_to_string(tensor), 'tf.int32:[[1, 2], [3, 4]]')

  @parameterized.named_parameters(
      ('int', 123, '123'),
      ('float', 1.23, '1.23'),
      ('bool', False, 'False'),
      ('string', 'abc', "'abc'"))
  def test_object_to_string_primitive(self, primitive, expected_result):
    self.assertEqual(
        tf_coder_utils.object_to_string(primitive), expected_result)

  def test_object_to_string_dtype(self):
    self.assertEqual(tf_coder_utils.object_to_string(tf.int32), 'tf.int32')

  def test_object_to_string_sequence(self):
    # `Named` is a class and should be capitalized. It is only used in this test
    # so it is declared here, not in the global scope.
    Named = collections.namedtuple('Named', ('a', 'b'))  # pylint: disable=invalid-name
    sequence = [123, tf.constant([1, 2]), (), Named(a=False, b=1.5)]
    self.assertEqual(
        tf_coder_utils.object_to_string(sequence),
        'seq[123, tf.int32:[1, 2], seq[], seq[False, 1.5]]')

  def test_object_to_string_raises_if_unsupported(self):
    with self.assertRaises(ValueError):
      tf_coder_utils.object_to_string({'key': 'value'})

  @parameterized.named_parameters(
      ('0_elements_1_part', 0, 1, [[0]]),
      ('0_elements_2_parts', 0, 2, [[0, 0]]),
      ('1_element_1_part', 1, 1, [[1]]),
      ('1_element_2_parts', 1, 2, [[0, 1], [1, 0]]),
      ('3_element_3_parts', 3, 3, [
          [0, 0, 3], [0, 1, 2], [0, 2, 1], [0, 3, 0],
          [1, 0, 2], [1, 1, 1], [1, 2, 0],
          [2, 0, 1], [2, 1, 0],
          [3, 0, 0],
      ]))
  def test_generate_partitions(self, num_elements, num_parts, expected_result):
    actual = list(tf_coder_utils.generate_partitions(num_elements, num_parts))
    self.assertCountEqual(actual, expected_result)

  @parameterized.named_parameters(
      ('negative_elements', -1, 1),
      ('negative_parts', 1, -1),
      ('zero_parts', 1, 0))
  def test_generate_partitions_raises_on_invalid_input(self, num_elements,
                                                       num_parts):
    with self.assertRaises(ValueError):
      list(tf_coder_utils.generate_partitions(num_elements, num_parts))


if __name__ == '__main__':
  logging.set_verbosity(logging.ERROR)

  absltest.main()
